@@ -144,13 +144,17 @@ val intern_constr_pattern :
   env -> evar_map -> ?as_type:bool -> ?strict_check:bool -> ?ltacvars:ltac_sign ->
     constr_pattern_expr -> Id.Set.t * constr_pattern
 
+val intern_uninstantiated_constr_pattern :
+  env -> evar_map -> ?as_type:bool -> ?strict_check:bool -> ?ltacvars:ltac_sign ->
+    constr_pattern_expr -> Id.Set.t * [`uninstantiated] constr_pattern_r
+
 (** Returns None if it's an abbreviation not bound to a name, raises an error
     if not existing *)
 val intern_reference : qualid -> GlobRef.t option
 
 (** Returns None if not a reference or a abbrev not bound to a name *)
 val intern_name_alias :
-  constr_expr -> (GlobRef.t * Glob_term.glob_level list option) option
+  constr_expr -> (GlobRef.t * Glob_term.glob_instance option) option
 
 (** Expands abbreviations; raise an error if not existing *)
 val interp_reference : ltac_sign -> qualid -> glob_constr
@@ -186,7 +190,7 @@ val is_global : Id.t -> bool
     guaranteed to have the same domain as the input one. *)
 val interp_notation_constr : env -> ?impls:internalization_env ->
   notation_interp_env -> constr_expr ->
-  (bool * subscopes) Id.Map.t * notation_constr * reversibility_status
+  (bool * subscopes * Id.Set.t) Id.Map.t * notation_constr * reversibility_status
 
 (** Idem but to glob_constr (weaker check of binders) *)
 
