@@ -15,12 +15,19 @@ open Names
 open Nameops
 open Constr
 open Globnames
-open Namegen
 open Glob_term
 open Glob_ops
 open Mod_subst
 open Notation_term
 
+(** Constr default entry *)
+
+let constr_lowest_level =
+  Constrexpr.{notation_entry = InConstrEntry; notation_level = 0}
+
+let constr_some_level =
+  Constrexpr.{notation_subentry = InConstrEntry; notation_relative_level = LevelSome; notation_position = None}
+
 (**********************************************************************)
 (* Utilities                                                          *)
 
@@ -64,6 +71,13 @@ let rec alpha_var id1 id2 = function
    in NList and NBinderList, since the iterator has its own variable *)
 let replace_var i j var = j :: List.remove Id.equal i var
 
+let eq_rigid a b =
+  let open UState in
+  match a, b with
+  | UnivRigid, UnivRigid -> true
+  | UnivFlexible a, UnivFlexible b -> (a:bool) = b
+  | (UnivRigid | UnivFlexible _), _ -> false
+
 (* compare_glob_universe_instances true strictly_lt us1 us2 computes us1 <= us2,
    compare_glob_universe_instances false strictly_lt us1 us2 computes us1 = us2.
    strictly_lt will be set to true if any part is strictly less. *)
@@ -72,15 +86,24 @@ let compare_glob_universe_instances lt strictly_lt us1 us2 =
   | None, None -> true
   | Some _, None -> strictly_lt := true; lt
   | None, Some _ -> false
-  | Some l1, Some l2 ->
-     CList.for_all2eq (fun u1 u2 ->
+  | Some (ql1,ul1), Some (ql2,ul2) ->
+    let is_anon = function
+      | GQualVar (GLocalQVar {v=Anonymous}) -> true
+      | _ -> false
+    in
+    CList.for_all2eq (fun q1 q2 ->
+        match is_anon q1, is_anon q2 with
+        | true, true -> true
+        | false, true -> strictly_lt := true; lt
+        | true, false -> false
+        | false, false -> glob_quality_eq q1 q2)
+      ql1 ql2
+    && CList.for_all2eq (fun u1 u2 ->
          match u1, u2 with
-         | UAnonymous {rigid=true}, UAnonymous {rigid=true} -> true
-         | UAnonymous {rigid=false}, UAnonymous {rigid=false} -> true
-         | UAnonymous _, UAnonymous _ -> false
+         | UAnonymous {rigid}, UAnonymous {rigid=rigid'} -> eq_rigid rigid rigid'
          | UNamed _, UAnonymous _ -> strictly_lt := true; lt
          | UAnonymous _, UNamed _ -> false
-         | UNamed _, UNamed _ -> glob_level_eq u1 u2) l1 l2
+         | UNamed _, UNamed _ -> glob_level_eq u1 u2) ul1 ul2
 
 (* Compute us1 <= us2, as a boolean *)
 let compare_glob_universe_instances_le us1 us2 =
@@ -137,8 +160,8 @@ let compare_notation_constr lt var_eq_hole (vars1,vars2) t1 t2 =
   | NVar id1, NHole _ when lt && List.mem_f Id.equal id1 vars1 -> ()
   | _, NVar id2 when lt && List.mem_f Id.equal id2 vars2 -> strictly_lt := true
   | NRef (gr1,u1), NRef (gr2,u2) when GlobRef.CanOrd.equal gr1 gr2 && compare_glob_universe_instances lt strictly_lt u1 u2 -> ()
-  | NHole (_, _), NHole (_, _) -> () (* FIXME? *)
-  | _, NHole (_, _) when lt -> strictly_lt := true
+  | NHole _, NHole _ -> () (* FIXME? *)
+  | _, NHole _ when lt -> strictly_lt := true
   | NGenarg _, NGenarg _ -> () (* FIXME? *)
   | NList (i1, j1, iter1, tail1, b1), NList (i2, j2, iter2, tail2, b2)
   | NBinderList (i1, j1, iter1, tail1, b1), NBinderList (i2, j2, iter2, tail2, b2) ->
@@ -303,27 +326,22 @@ let rec cases_pattern_fold_map ?loc g e = DAst.with_val (function
   )
 
 let subst_binder_type_vars l = function
-  | Evar_kinds.BinderType (Name id) ->
+  | GBinderType (Name id) ->
      let id =
        try match DAst.get (Id.List.assoc id l) with GVar id' -> id' | _ -> id
        with Not_found -> id in
-     Evar_kinds.BinderType (Name id)
+     GBinderType (Name id)
   | e -> e
 
 let rec subst_glob_vars l gc = DAst.map (function
   | GVar id as r -> (try DAst.get (Id.List.assoc id l) with Not_found -> r)
-  | GProd (Name id,bk,t,c) ->
-      let id =
-        try match DAst.get (Id.List.assoc id l) with GVar id' -> id' | _ -> id
-        with Not_found -> id in
-      GProd (Name id,bk,subst_glob_vars l t,subst_glob_vars l c)
-  | GLambda (Name id,bk,t,c) ->
-      let id =
-        try match DAst.get (Id.List.assoc id l) with GVar id' -> id' | _ -> id
-        with Not_found -> id in
-      GLambda (Name id,bk,subst_glob_vars l t,subst_glob_vars l c)
-  | GHole (x,naming) -> GHole (subst_binder_type_vars l x,naming)
-  | _ -> DAst.get (map_glob_constr (subst_glob_vars l) gc) (* assume: id is not binding *)
+  | GHole x -> GHole (subst_binder_type_vars l x)
+  | _ ->
+    let g =
+      Name.map (fun id ->
+          try match DAst.get (Id.List.assoc id l) with GVar id' -> id' | _ -> id
+          with Not_found -> id) in
+    DAst.get (map_glob_constr_left_to_right_with_names (subst_glob_vars l) g gc) (* assume: id is not binding *)
   ) gc
 
 type 'a binder_status_fun = {
@@ -360,7 +378,7 @@ let protect g e na =
   e',na
 
 let set_anonymous_type na = function
-  | None -> DAst.make @@ GHole (Evar_kinds.BinderType na, IntroAnonymous)
+  | None -> DAst.make @@ GHole (GBinderType na)
   | Some t -> t
 
 let apply_cases_pattern_term ?loc (ids,disjpat) tm c =
@@ -445,7 +463,7 @@ let glob_constr_of_notation_constr_with_binders ?loc g f ?(h=default_binder_stat
       GRec (fk,idl,dll,Array.map (f e) tl,Array.map (f e') bl)
   | NCast (c,k,t) -> GCast (f e c, k, f (h.slide e) t)
   | NSort x -> GSort x
-  | NHole (x, naming)  -> GHole (x, naming)
+  | NHole x  -> GHole x
   | NGenarg arg -> GGenarg arg
   | NRef (x,u) -> GRef (x,u)
   | NInt i -> GInt i
@@ -457,6 +475,10 @@ let glob_constr_of_notation_constr ?loc x =
     glob_constr_of_notation_constr_with_binders ?loc (fun () id t -> ((),None,id,Explicit,t)) aux () x
   in aux () x
 
+let pr_notation_info prglob ntn c =
+  str (String.quote_coq_string ntn) ++ str " :=" ++ brk (1,2) ++
+  prglob (glob_constr_of_notation_constr c)
+
 (******************************************************************************)
 (* Translating a glob_constr into a notation, interpreting recursive patterns *)
 
@@ -513,7 +535,9 @@ let subtract_loc loc1 loc2 =
   let l2 = fst (Option.cata Loc.unloc (0,0) loc2) in
   Some (Loc.make_loc (l1,l2-1))
 
-let check_is_hole id t = match DAst.get t with GHole _ -> () | _ ->
+let check_is_hole id = function
+  | None -> ()
+  | Some t -> match DAst.get t with GHole _ -> () | _ ->
   user_err ?loc:(loc_of_glob_constr t)
    (strbrk "In recursive notation with binders, " ++ Id.print id ++
     strbrk " is expected to come without type.")
@@ -528,9 +552,7 @@ let check_pair_matching ?loc x y x' y' revert revert' =
        strbrk " while " ++ Id.print x' ++ strbrk " matching " ++ Id.print y' ++
        strbrk " was first found.")
 
-let pair_equal eq1 eq2 (a,b) (a',b') = eq1 a a' && eq2 b b'
-
-let mem_recursive_pair (x,y) l = List.mem_f (pair_equal Id.equal Id.equal) (x,y) l
+let mem_recursive_pair (x,y) l = List.mem_f (eq_pair Id.equal Id.equal) (x,y) l
 
 type recursive_pattern_kind =
 | RecursiveTerms of bool (* in reverse order *)
@@ -539,6 +561,27 @@ type recursive_pattern_kind =
 let compare_recursive_parts recvars found f f' (iterator,subc) =
   let diff = ref None in
   let terminator = ref None in
+  let treat_binder ?loc x y t_x t_y =
+    match x, y with
+    | Name x, Name y when mem_recursive_pair (x,y) recvars || mem_recursive_pair (y,x) recvars ->
+      (* We found a binding position where it differs *)
+      check_is_hole x t_x;
+      check_is_hole y t_y;
+      let revert = mem_recursive_pair (y,x) recvars in
+      let x,y = if revert then y,x else x,y in
+      begin match !diff with
+      | None -> diff := Some (x, y, RecursiveBinders revert)
+      | Some (x', y', RecursiveBinders revert') ->
+        check_pair_matching ?loc x y x' y' revert revert'
+      | Some (x', y', RecursiveTerms revert') ->
+        (* Recursive binders have precedence: they can be coerced to
+           terms but not reciprocally *)
+        check_pair_matching ?loc x y x' y' revert revert';
+        diff := Some (x, y, RecursiveBinders revert)
+      end;
+      true
+    | _ -> Name.equal x y
+  in
   let rec aux c1 c2 = match DAst.get c1, DAst.get c2 with
   | GVar v, term when Id.equal v ldots_var ->
       (* We found the pattern *)
@@ -553,7 +596,7 @@ let compare_recursive_parts recvars found f f' (iterator,subc) =
       assert (match !terminator with None -> true | Some _ -> false);
       terminator := Some term;
       List.for_all2eq aux l1 l2
-    | _ -> mk_glob_constr_eq aux c1 c2
+    | _ -> mk_glob_constr_eq aux (treat_binder ?loc:c1.CAst.loc) c1 c2
     end
   | GVar x, GVar y
         when mem_recursive_pair (x,y) recvars || mem_recursive_pair (y,x) recvars ->
@@ -569,30 +612,8 @@ let compare_recursive_parts recvars found f f' (iterator,subc) =
         check_pair_matching ?loc:c1.CAst.loc x y x' y' revert revert';
         true
       end
-  | GLambda (Name x,_,t_x,c), GLambda (Name y,_,t_y,term)
-  | GProd (Name x,_,t_x,c), GProd (Name y,_,t_y,term)
-        when mem_recursive_pair (x,y) recvars || mem_recursive_pair (y,x) recvars ->
-      (* We found a binding position where it differs *)
-      check_is_hole x t_x;
-      check_is_hole y t_y;
-      let revert = mem_recursive_pair (y,x) recvars in
-      let x,y = if revert then y,x else x,y in
-      begin match !diff with
-      | None ->
-        let () = diff := Some (x, y, RecursiveBinders revert) in
-        aux c term
-      | Some (x', y', RecursiveBinders revert') ->
-        check_pair_matching ?loc:c1.CAst.loc x y x' y' revert revert';
-        true
-      | Some (x', y', RecursiveTerms revert') ->
-        (* Recursive binders have precedence: they can be coerced to
-           terms but not reciprocally *)
-        check_pair_matching ?loc:c1.CAst.loc x y x' y' revert revert';
-        let () = diff := Some (x, y, RecursiveBinders revert) in
-        true
-      end
   | _ ->
-      mk_glob_constr_eq aux c1 c2 in
+      mk_glob_constr_eq aux (treat_binder ?loc:c1.CAst.loc) c1 c2 in
   if aux iterator subc then
     match !diff with
     | None ->
@@ -608,14 +629,14 @@ let compare_recursive_parts recvars found f f' (iterator,subc) =
           f' (if revert then iterator else subst_glob_vars [x, DAst.make @@ GVar y] iterator) in
         (* found variables have been collected by compare_constr *)
         found := { !found with vars = List.remove Id.equal y (!found).vars;
-                               recursive_term_vars = List.add_set (pair_equal Id.equal Id.equal) (x,y) (!found).recursive_term_vars };
+                               recursive_term_vars = List.add_set (eq_pair Id.equal Id.equal) (x,y) (!found).recursive_term_vars };
         NList (x,y,iterator,f (Option.get !terminator),revert)
     | Some (x,y,RecursiveBinders revert) ->
         let iterator =
           f' (if revert then iterator else subst_glob_vars [x, DAst.make @@ GVar y] iterator) in
         (* found have been collected by compare_constr *)
         found := { !found with vars = List.remove Id.equal y (!found).vars;
-                               recursive_binders_vars = List.add_set (pair_equal Id.equal Id.equal) (x,y) (!found).recursive_binders_vars };
+                               recursive_binders_vars = List.add_set (eq_pair Id.equal Id.equal) (x,y) (!found).recursive_binders_vars };
         NBinderList (x,y,iterator,f (Option.get !terminator),revert)
   else
     raise Not_found
@@ -680,7 +701,7 @@ let notation_constr_and_vars_of_glob_constr recvars a =
   | GSort s -> NSort s
   | GInt i -> NInt i
   | GFloat f -> NFloat f
-  | GHole (w,naming) -> NHole (w, naming)
+  | GHole w -> NHole w
   | GGenarg arg -> has_ltac := true; NGenarg arg
   | GRef (r,u) -> NRef (r,u)
   | GArray (_u,t,def,ty) -> NArray (Array.map aux t, aux def, aux ty)
@@ -688,7 +709,7 @@ let notation_constr_and_vars_of_glob_constr recvars a =
       user_err Pp.(str "Existential variables not allowed in notations.")
   ) x
   and aux_type t = DAst.with_val (function
-  | GHole (Evar_kinds.BinderType _,IntroAnonymous) -> None
+  | GHole (GBinderType _) -> None
   | _ -> Some (aux t)) t
   in
   let t = aux a in
@@ -781,7 +802,7 @@ let rec subst_notation_constr subst bound raw =
       if ref' == ref then raw else (match t with
           | None -> NRef (ref',u)
           | Some t ->
-            fst (notation_constr_of_constr bound t.Univ.univ_abstracted_value))
+            fst (notation_constr_of_constr bound t.UVars.univ_abstracted_value))
 
   | NVar _ -> raw
 
@@ -883,18 +904,18 @@ let rec subst_notation_constr subst bound raw =
   | NInt _ -> raw
   | NFloat _ -> raw
 
-  | NHole (knd, naming) ->
+  | NHole knd ->
     let nknd = match knd with
-    | Evar_kinds.ImplicitArg (ref, i, b) ->
+    | GImplicitArg (ref, i, b) ->
       let nref, _ = subst_global subst ref in
-      if nref == ref then knd else Evar_kinds.ImplicitArg (nref, i, b)
+      if nref == ref then knd else GImplicitArg (nref, i, b)
     | _ -> knd
     in
     if nknd == knd then raw
-    else NHole (nknd, naming)
+    else NHole nknd
 
   | NGenarg arg ->
-    let arg' = Genintern.generic_substitute subst arg in
+    let arg' = Gensubst.generic_substitute subst arg in
     if arg' == arg then raw
     else NGenarg arg'
 
@@ -929,7 +950,7 @@ let abstract_return_type_context pi mklam tml rtno =
 let abstract_return_type_context_glob_constr tml rtn =
   abstract_return_type_context (fun {CAst.v=(_,nal)} -> nal)
     (fun na c -> DAst.make @@
-      GLambda(na,Explicit,DAst.make @@ GHole(Evar_kinds.InternalHole,IntroAnonymous),c)) tml rtn
+      GLambda(na,Explicit,DAst.make @@ GHole (GInternalHole), c)) tml rtn
 
 let abstract_return_type_context_notation_constr tml rtn =
   abstract_return_type_context snd
@@ -950,29 +971,33 @@ let rec push_context_binders vars = function
     push_context_binders vars bl
 
 let is_term_meta id metas =
-  try match Id.List.assoc id metas with _,(NtnTypeConstr | NtnTypeConstrList) -> true | _ -> false
+  try match Id.List.assoc id metas with NtnTypeConstr | NtnTypeConstrList -> true | _ -> false
   with Not_found -> false
 
 let is_onlybinding_strict_meta id metas =
-  try match Id.List.assoc id metas with _,NtnTypeBinder (NtnBinderParsedAsSomeBinderKind AsStrictPattern) -> true | _ -> false
+  try match Id.List.assoc id metas with NtnTypeBinder (NtnBinderParsedAsSomeBinderKind AsStrictPattern) -> true | _ -> false
   with Not_found -> false
 
 let is_onlybinding_meta id metas =
-  try match Id.List.assoc id metas with _,NtnTypeBinder _ -> true | _ -> false
+  try match Id.List.assoc id metas with NtnTypeBinder _ -> true | _ -> false
+  with Not_found -> false
+
+let is_onlybindinglist_meta id metas =
+  try match Id.List.assoc id metas with NtnTypeBinderList _ -> true | _ -> false
   with Not_found -> false
 
 let is_onlybinding_pattern_like_meta isvar id metas =
   try match Id.List.assoc id metas with
-    | _,NtnTypeBinder (NtnBinderParsedAsConstr (AsAnyPattern | AsStrictPattern)) -> true
-    | _,NtnTypeBinder (NtnBinderParsedAsSomeBinderKind AsStrictPattern | NtnBinderParsedAsBinder) -> not isvar
-    | _,NtnTypeBinder (NtnBinderParsedAsSomeBinderKind AsAnyPattern) -> true
-    | _,NtnTypeBinder (NtnBinderParsedAsSomeBinderKind (AsIdent | AsName)) -> false
-    | _,NtnTypeBinder (NtnBinderParsedAsConstr (AsIdent | AsName)) -> false
-    | _,NtnTypeBinderList _ | _,NtnTypeConstr | _,NtnTypeConstrList -> false
+    | NtnTypeBinder (NtnBinderParsedAsConstr (AsAnyPattern | AsStrictPattern)) -> true
+    | NtnTypeBinder (NtnBinderParsedAsSomeBinderKind AsStrictPattern | NtnBinderParsedAsBinder) -> not isvar
+    | NtnTypeBinder (NtnBinderParsedAsSomeBinderKind AsAnyPattern) -> true
+    | NtnTypeBinder (NtnBinderParsedAsSomeBinderKind (AsIdent | AsName)) -> false
+    | NtnTypeBinder (NtnBinderParsedAsConstr (AsIdent | AsName)) -> false
+    | NtnTypeBinderList _ | NtnTypeConstr | NtnTypeConstrList -> false
   with Not_found -> false
 
 let is_bindinglist_meta id metas =
-  try match Id.List.assoc id metas with _,NtnTypeBinderList _ -> true | _ -> false
+  try match Id.List.assoc id metas with NtnTypeBinderList _ -> true | _ -> false
   with Not_found -> false
 
 exception No_match
@@ -1293,8 +1318,17 @@ let rec match_cases_pattern_binders allow_catchall metas (alp,sigma as acc) pat1
   match DAst.get pat1, DAst.get pat2 with
   | PatVar _, PatVar (Name id2) when is_onlybinding_pattern_like_meta true id2 metas ->
       bind_binding_env alp sigma id2 [pat1]
+  | PatVar id1, PatVar (Name id2) when is_onlybindinglist_meta id2 metas ->
+      let t1 = DAst.make @@ GHole(GBinderType id1) in
+      bind_bindinglist_env alp sigma id2 [DAst.make @@ GLocalAssum (id1,Explicit,t1)]
   | _, PatVar (Name id2) when is_onlybinding_pattern_like_meta false id2 metas ->
       bind_binding_env alp sigma id2 [pat1]
+  | _, PatVar (Name id2) when is_onlybindinglist_meta id2 metas ->
+      (* dummy data; should not be used anyway *)
+      let id1 = Namegen.next_ident_away (Id.of_string "x") Id.Set.empty in
+      let t1 = DAst.make @@ GHole(GBinderType (Name id1)) in
+      let ids1 = [] in
+      bind_bindinglist_env alp sigma id2 [DAst.make @@ GLocalPattern (([pat1],ids1),id1,Explicit,t1)]
   | PatVar na1, PatVar na2 -> match_names metas acc na1 na2
   | _, PatVar Anonymous when allow_catchall -> acc
   | PatCstr (c1,patl1,na1), PatCstr (c2,patl2,na2)
@@ -1309,11 +1343,9 @@ let remove_sigma x (terms,termlists,binders,binderlists) =
 let remove_bindinglist_sigma x (terms,termlists,binders,binderlists) =
   (terms,termlists,binders,Id.List.remove_assoc x binderlists)
 
-let default_constr_entry_relative_level = Constrexpr.(InConstrEntry,(LevelSome,None))
+let add_ldots_var metas = (ldots_var,NtnTypeConstr)::metas
 
-let add_ldots_var metas = (ldots_var,((default_constr_entry_relative_level,([],[])),NtnTypeConstr))::metas
-
-let add_meta_bindinglist x metas = (x,((default_constr_entry_relative_level,([],[])),NtnTypeBinderList (*arbitrary:*) NtnBinderParsedAsBinder))::metas
+let add_meta_bindinglist x metas = (x,NtnTypeBinderList (*arbitrary:*) NtnBinderParsedAsBinder)::metas
 
 (* This tells if letins in the middle of binders should be included in
    the sequence of binders *)
@@ -1359,7 +1391,7 @@ let match_binderlist match_iter_fun match_termin_fun alp metas sigma rest x y it
   let alp,sigma = bind_bindinglist_env alp sigma x bl in
   match_termin_fun alp metas sigma rest termin
 
-let add_meta_term x metas = (x,((default_constr_entry_relative_level,([],[])),NtnTypeConstr))::metas (* Should reuse the scope of the partner of x! *)
+let add_meta_term x metas = (x,NtnTypeConstr)::metas
 
 let match_termlist match_fun alp metas sigma rest x y iter termin revert =
   let rec aux alp sigma acc rest =
@@ -1397,6 +1429,11 @@ let does_not_come_from_already_eta_expanded_var glob =
   (* checked). *)
   match DAst.get glob with GVar _ -> false | _ -> true
 
+let eta_well_typed pat =
+  (* A criterion for well-typedness of the eta-expansion
+     is that the head of the pattern is rigid (see #15221) *)
+  match pat with NVar _ -> false | _ -> true
+
 let is_var_term = function
   (* The kind of expressions allowed to be both a term and a binding variable *)
   | GVar _ -> true
@@ -1440,7 +1477,7 @@ let rec match_ inner u alp metas sigma a1 a2 =
         else if n1 > n2 then
           let l11,l12 = List.chop (n1-n2) l1 in DAst.make ?loc @@ GApp (f1,l11),l12, f2,l2
         else f1,l1, f2, l2 in
-      let may_use_eta = does_not_come_from_already_eta_expanded_var f1 in
+      let may_use_eta = does_not_come_from_already_eta_expanded_var f1 && eta_well_typed f2 in
       List.fold_left2 (match_ may_use_eta u alp metas)
         (match_hd u alp metas sigma f1 f2) l1 l2
   | GProj ((cst1,u1),l1,a1), NProj ((cst2,u2),l2,a2) when GlobRef.CanOrd.equal (GlobRef.ConstRef cst1) (GlobRef.ConstRef cst2) && compare_glob_universe_instances_le u1 u2 ->
@@ -1455,11 +1492,12 @@ let rec match_ inner u alp metas sigma a1 a2 =
      match_extended_binders false u alp metas na1 na2 bk1 t1 (match_in_type u alp metas sigma t1 t2) b1 b2
   | GProd (na1,bk1,t1,b1), NProd (na2,t2,b2) ->
      match_extended_binders (not inner) u alp metas na1 na2 bk1 t1 (match_in_type u alp metas sigma t1 t2) b1 b2
-  | GLetIn (na1,b1,_,c1), NLetIn (na2,b2,None,c2)
-  | GLetIn (na1,b1,None,c1), NLetIn (na2,b2,_,c2) ->
-     match_binders u alp metas na1 na2 (match_in u alp metas sigma b1 b2) c1 c2
+  | GLetIn (na1,b1,t1,c1), NLetIn (na2,b2,None,c2)
+  | GLetIn (na1,b1,(None as t1),c1), NLetIn (na2,b2,_,c2) ->
+     let t = match t1 with Some t -> t | None -> DAst.make @@ GHole(GBinderType na1) in
+     match_extended_binders false u alp metas na1 na2 Explicit t (match_in u alp metas sigma b1 b2) c1 c2
   | GLetIn (na1,b1,Some t1,c1), NLetIn (na2,b2,Some t2,c2) ->
-     match_binders u alp metas na1 na2
+     match_extended_binders false u alp metas na1 na2 Explicit t1
        (match_in u alp metas (match_in u alp metas sigma b1 b2) t1 t2) c1 c2
   | GCases (sty1,rtno1,tml1,eqnl1), NCases (sty2,rtno2,tml2,eqnl2)
       when sty1 == sty2 && Int.equal (List.length tml1) (List.length tml2) ->
@@ -1529,7 +1567,7 @@ let rec match_ inner u alp metas sigma a1 a2 =
       let avoid =
         Id.Set.union (free_glob_vars a1) (* as in Namegen: *) (glob_visible_short_qualid a1) in
       let id' = Namegen.next_ident_away id avoid in
-      let t1 = DAst.make @@ GHole(Evar_kinds.BinderType (Name id'),IntroAnonymous) in
+      let t1 = DAst.make @@ GHole (GBinderType (Name id')) in
       let sigma = match t2 with
       | None -> sigma
       | Some (NVar id2) -> bind_term_env alp sigma id2 t1
@@ -1619,7 +1657,7 @@ and match_disjunctive_equations u alp metas sigma {CAst.v=(ids,disjpatl1,rhs1)}
    substitution based on entry production: indeed some binders may
    have to be seen as terms from the parsing/printing point of view *)
 let group_by_type ids (terms,termlists,binders,binderlists) =
-  List.fold_right (fun (x,(scl,typ)) (terms',termlists',binders',binderlists') ->
+  List.fold_right (fun (x,(scl,_,typ)) (terms',termlists',binders',binderlists') ->
     match typ with
     | NtnTypeConstr ->
        (* term -> term *)
@@ -1645,7 +1683,7 @@ let group_by_type ids (terms,termlists,binders,binderlists) =
            | GLocalPattern ((disjpat,_),_,_,_) ->
              List.map (glob_constr_of_cases_pattern (Global.env())) disjpat
            | GLocalAssum (Anonymous,bk,t) ->
-             let hole = DAst.make (GHole (Evar_kinds.BinderType Anonymous,IntroAnonymous)) in
+             let hole = DAst.make (GHole (GBinderType Anonymous)) in
              [DAst.make (GCast (hole, Some DEFAULTcast, t))]
            | GLocalAssum (Name id,bk,t) ->
              [DAst.make (GCast (DAst.make (GVar id), Some DEFAULTcast, t))]
@@ -1658,7 +1696,8 @@ let group_by_type ids (terms,termlists,binders,binderlists) =
     ids ([],[],[],[])
 
 let match_notation_constr ~print_univ c ~vars (metas,pat) =
-  let subst = match_ false print_univ {actualvars=vars;staticbinders=[];renaming=[]} metas ([],[],[],[]) c pat in
+  let metatyps = List.map (fun (id,(_,_,typ)) -> (id,typ)) metas in
+  let subst = match_ false print_univ {actualvars=vars;staticbinders=[];renaming=[]} metatyps ([],[],[],[]) c pat in
   group_by_type metas subst
 
 (* Matching cases pattern *)
@@ -1731,17 +1770,21 @@ let match_ind_pattern metas sigma ind pats a2 =
   |_ -> raise No_match
 
 let reorder_canonically_substitution terms termlists metas =
-  List.fold_right (fun (x,(scl,typ)) (terms',termlists') ->
+  List.fold_right (fun (x,(scl,_,typ)) (terms',termlists',binders') ->
     match typ with
-      | NtnTypeConstr -> ((Id.List.assoc x terms, scl)::terms',termlists')
-      | NtnTypeConstrList -> (terms',(Id.List.assoc x termlists,scl)::termlists')
-      | NtnTypeBinder _ | NtnTypeBinderList _ -> anomaly (str "Unexpected binder in pattern notation."))
-    metas ([],[])
+      | NtnTypeConstr | NtnTypeBinder (NtnBinderParsedAsConstr _) -> ((Id.List.assoc x terms, scl)::terms',termlists',binders')
+      | NtnTypeConstrList -> (terms',(Id.List.assoc x termlists,scl)::termlists',binders')
+      | NtnTypeBinder (NtnBinderParsedAsBinder | NtnBinderParsedAsSomeBinderKind _) ->
+         (terms',termlists',(Id.List.assoc x terms, scl)::binders')
+      | NtnTypeBinderList _ -> anomaly (str "Unexpected binder in pattern notation."))
+    metas ([],[],[])
 
 let match_notation_constr_cases_pattern c (metas,pat) =
-  let (terms,termlists,(),()),more_args = match_cases_pattern metas ([],[],(),()) c pat in
+  let metatyps = List.map (fun (id,(_,_,typ)) -> (id,typ)) metas in
+  let (terms,termlists,(),()),more_args = match_cases_pattern metatyps ([],[],(),()) c pat in
   reorder_canonically_substitution terms termlists metas, more_args
 
 let match_notation_constr_ind_pattern ind args (metas,pat) =
-  let (terms,termlists,(),()),more_args = match_ind_pattern metas ([],[],(),()) ind args pat in
+  let metatyps = List.map (fun (id,(_,_,typ)) -> (id,typ)) metas in
+  let (terms,termlists,(),()),more_args = match_ind_pattern metatyps ([],[],(),()) ind args pat in
   reorder_canonically_substitution terms termlists metas, more_args
