@@ -73,9 +73,9 @@ sp is a local copy of the global variable extern_sp. */
 #ifdef THREADED_CODE
 #  define Instruct(name) coq_lbl_##name:
 #  if defined(ARCH_SIXTYFOUR) && !defined(ARCH_CODE32)
-#    define coq_Jumptbl_base ((char *) &&coq_lbl_ACC0)
+#    define coq_Jumptbl_base &&coq_lbl_ACC0
 #  else
-#    define coq_Jumptbl_base ((char *) 0)
+#    define coq_Jumptbl_base 0
 #    define coq_jumptbl_base ((char *) 0)
 #  endif
 #  ifdef DEBUG
@@ -91,9 +91,9 @@ sp is a local copy of the global variable extern_sp. */
 /* #define _COQ_DEBUG_ */
 
 #ifdef _COQ_DEBUG_
-#   define print_instr(s) /*if (drawinstr)*/ printf("%s\n",s)
-#   define print_int(i)   /*if (drawinstr)*/ printf("%d\n",i)
-#   define print_lint(i)  /*if (drawinstr)*/ printf("%ld\n",i)
+#   define print_instr(s) printf("%s\n",s)
+#   define print_int(i)   printf("%d\n",i)
+#   define print_lint(i)  printf("%ld\n",i)
 # else
 #   define print_instr(s)
 #   define print_int(i)
@@ -115,8 +115,12 @@ if (sp - num_args < coq_stack_threshold) {                                     \
 #define Restore_after_caml_call coq_env = *sp++;
 
 #if OCAML_VERSION >= 50000
-#define Coq_alloc_small(result, wosize, tag) Alloc_small(result, wosize, tag, \
-  { Setup_for_gc; caml_process_pending_actions(); Restore_after_gc; })
+#define Enter_gc(dom_st, wosize) do {        \
+    Setup_for_gc;                            \
+    Alloc_small_enter_GC(dom_st, wosize);    \
+    Restore_after_gc;                        \
+  } while (0)
+#define Coq_alloc_small(result, wosize, tag) Alloc_small(result, wosize, tag, Enter_gc)
 #else
 #define Coq_alloc_small(result, wosize, tag) Alloc_small(result, wosize, tag)
 #endif
@@ -297,8 +301,7 @@ value coq_interprete
   if (coq_pc == NULL) {           /* Interpreter is initializing */
     print_instr("Interpreter is initializing");
 #ifdef THREADED_CODE
-    coq_instr_table = (char **) coq_jumptable;
-    coq_instr_base = coq_Jumptbl_base;
+    coq_init_thread_code(coq_jumptable, coq_Jumptbl_base);
 #endif
     CAMLreturn(Val_unit);
   }
@@ -567,22 +570,27 @@ value coq_interprete
       CHECK_STACK(0);
       /* We also check for signals */
 #if OCAML_VERSION >= 50000
-      if (Caml_check_gc_interrupt(Caml_state) || caml_check_for_pending_signals()) {
-        // FIXME: it should be caml_process_pending_actions_exn
-        value res = caml_process_pending_signals_exn();
+      if (Caml_check_gc_interrupt(Caml_state)) {
+        Setup_for_gc;
+        value res = caml_process_pending_actions_exn();
         Handle_potential_exception(res);
+        Restore_after_gc;
       }
 #elif OCAML_VERSION >= 41000
       if (caml_something_to_do) {
+        Setup_for_gc;
         value res = caml_process_pending_actions_exn();
         Handle_potential_exception(res);
+        Restore_after_gc;
       }
 #else
       if (caml_signals_are_pending) {
         /* If there's a Ctrl-C, we reset the vm */
         intnat sigint = caml_pending_signals[SIGINT];
         if (sigint) { coq_sp = coq_stack_high; }
+        Setup_for_gc;
         caml_process_pending_signals();
+        Restore_after_gc;
         if (sigint) {
           caml_failwith("Coq VM: Fatal error: SIGINT signal detected "
                         "but no exception was raised");
@@ -2004,8 +2012,9 @@ value  coq_interprete_ml(value tcode, value a, value t, value g, value e, value
   // Registering the other arguments w.r.t. the OCaml GC is done by coq_interprete
   CAMLparam1(tcode);
   print_instr("coq_interprete");
-  CAMLreturn (coq_interprete(Code_val(tcode), a, t, g, e, Long_val(ea)));
+  value res = coq_interprete(Code_val(tcode), a, t, g, e, Long_val(ea));
   print_instr("end coq_interprete");
+  CAMLreturn(res);
 }
 
 value coq_interprete_byte(value* argv, int argn){
