@@ -63,28 +63,15 @@ module Globals = struct
   let view x = x
 end
 
-type val_kind =
-    | VKvalue of (Vmvalues.values * Id.Set.t) CEphemeron.key
-    | VKnone
-
-type lazy_val = val_kind ref
-
-let force_lazy_val vk = match !vk with
-| VKnone -> None
-| VKvalue v -> try Some (CEphemeron.get v) with CEphemeron.InvalidKey -> None
-
-let dummy_lazy_val () = ref VKnone
-let build_lazy_val vk key = vk := VKvalue (CEphemeron.create key)
-
 type named_context_val = {
   env_named_ctx : Constr.named_context;
-  env_named_map : (Constr.named_declaration * lazy_val) Id.Map.t;
-  env_named_var : Constr.t list;
+  env_named_map : Constr.named_declaration Id.Map.t;
+  env_named_idx : Constr.named_declaration Range.t;
 }
 
 type rel_context_val = {
   env_rel_ctx : Constr.rel_context;
-  env_rel_map : (Constr.rel_declaration * lazy_val) Range.t;
+  env_rel_map : Constr.rel_declaration Range.t;
 }
 
 type env = {
@@ -94,17 +81,17 @@ type env = {
   env_nb_rel        : int;
   env_universes : UGraph.t;
   env_universes_lbound : UGraph.Bound.t;
-  irr_constants : Cset_env.t;
-  irr_inds : Indset_env.t;
+  env_qualities : Sorts.QVar.Set.t;
+  irr_constants : Sorts.relevance Cmap_env.t;
+  irr_inds : Sorts.relevance Indmap_env.t;
   env_typing_flags  : typing_flags;
   retroknowledge : Retroknowledge.retroknowledge;
-  indirect_pterms : Opaqueproof.opaquetab;
 }
 
 let empty_named_context_val = {
   env_named_ctx = [];
   env_named_map = Id.Map.empty;
-  env_named_var = [];
+  env_named_idx = Range.empty;
 }
 
 let empty_rel_context_val = {
@@ -124,11 +111,11 @@ let empty_env = {
   env_nb_rel = 0;
   env_universes = UGraph.initial_universes;
   env_universes_lbound = UGraph.Bound.Set;
-  irr_constants = Cset_env.empty;
-  irr_inds = Indset_env.empty;
+  env_qualities = Sorts.QVar.Set.empty;
+  irr_constants = Cmap_env.empty;
+  irr_inds = Indmap_env.empty;
   env_typing_flags = Declareops.safe_flags Conv_oracle.empty;
   retroknowledge = Retroknowledge.empty;
-  indirect_pterms = Opaqueproof.empty_opaquetab;
 }
 
 
@@ -136,15 +123,14 @@ let empty_env = {
 
 let push_rel_context_val d ctx = {
   env_rel_ctx = Context.Rel.add d ctx.env_rel_ctx;
-  env_rel_map = Range.cons (d, ref VKnone) ctx.env_rel_map;
+  env_rel_map = Range.cons d ctx.env_rel_map;
 }
 
 let match_rel_context_val ctx = match ctx.env_rel_ctx with
 | [] -> None
 | decl :: rem ->
-  let (_, lval) = Range.hd ctx.env_rel_map in
   let ctx = { env_rel_ctx = rem; env_rel_map = Range.tl ctx.env_rel_map } in
-  Some (decl, lval, ctx)
+  Some (decl, ctx)
 
 let push_rel d env =
     { env with
@@ -152,11 +138,7 @@ let push_rel d env =
       env_nb_rel = env.env_nb_rel + 1 }
 
 let lookup_rel n env =
-  try fst (Range.get env.env_rel_context.env_rel_map (n - 1))
-  with Invalid_argument _ -> raise Not_found
-
-let lookup_rel_val n env =
-  try snd (Range.get env.env_rel_context.env_rel_map (n - 1))
+  try Range.get env.env_rel_context.env_rel_map (n - 1)
   with Invalid_argument _ -> raise Not_found
 
 let rel_skipn n ctx = {
@@ -170,26 +152,27 @@ let env_of_rel n env =
     env_nb_rel = env.env_nb_rel - n
   }
 
+let set_rel_context_val v env =
+  { env with
+    env_rel_context = v;
+    env_nb_rel = Range.length v.env_rel_map; }
+
 (* Named context *)
 
-let push_named_context_val_val d rval ctxt =
+let push_named_context_val d ctxt =
 (*   assert (not (Id.Map.mem (NamedDecl.get_id d) ctxt.env_named_map)); *)
   {
     env_named_ctx = Context.Named.add d ctxt.env_named_ctx;
-    env_named_map = Id.Map.add (NamedDecl.get_id d) (d, rval) ctxt.env_named_map;
-    env_named_var = mkVar (NamedDecl.get_id d) :: ctxt.env_named_var;
+    env_named_map = Id.Map.add (NamedDecl.get_id d) d ctxt.env_named_map;
+    env_named_idx = Range.cons d ctxt.env_named_idx;
   }
 
-let push_named_context_val d ctxt =
-  push_named_context_val_val d (ref VKnone) ctxt
-
 let match_named_context_val c = match c.env_named_ctx with
 | [] -> None
 | decl :: ctx ->
-  let (_, v) = Id.Map.find (NamedDecl.get_id decl) c.env_named_map in
   let map = Id.Map.remove (NamedDecl.get_id decl) c.env_named_map in
-  let cval = { env_named_ctx = ctx; env_named_map = map; env_named_var = List.tl c.env_named_var } in
-  Some (decl, v, cval)
+  let cval = { env_named_ctx = ctx; env_named_map = map; env_named_idx = Range.tl c.env_named_idx } in
+  Some (decl, cval)
 
 let map_named_val f ctxt =
   let open Context.Named.Declaration in
@@ -197,25 +180,24 @@ let map_named_val f ctxt =
     let d' = f d in
     let accu =
       if d == d' then accu
-      else Id.Map.modify (get_id d) (fun _ (_, v) -> (d', v)) accu
+      else Id.Map.set (get_id d) d' accu
     in
     (accu, d')
   in
   let map, ctx = List.Smart.fold_left_map fold ctxt.env_named_map ctxt.env_named_ctx in
   if map == ctxt.env_named_map then ctxt
-  else { env_named_ctx = ctx; env_named_map = map; env_named_var = ctxt.env_named_var }
+  else
+    let idx = List.fold_right Range.cons ctx Range.empty in
+    { env_named_ctx = ctx; env_named_map = map; env_named_idx = idx }
 
 let push_named d env =
   {env with env_named_context = push_named_context_val d env.env_named_context}
 
 let lookup_named id env =
-  fst (Id.Map.find id env.env_named_context.env_named_map)
-
-let lookup_named_val id env =
-  snd(Id.Map.find id env.env_named_context.env_named_map)
+  Id.Map.find id env.env_named_context.env_named_map
 
 let lookup_named_ctxt id ctxt =
-  fst (Id.Map.find id ctxt.env_named_map)
+  Id.Map.find id ctxt.env_named_map
 
 let fold_constants f env acc =
   Cmap_env.fold (fun c (body,_) acc -> f c body acc) env.env_globals.Globals.constants acc
@@ -264,12 +246,12 @@ let is_impredicative_set env = env.env_typing_flags.impredicative_set
 let is_impredicative_sort env = function
   | Sorts.SProp | Sorts.Prop -> true
   | Sorts.Set -> is_impredicative_set env
-  | Sorts.Type _ -> false
+  | Sorts.Type _ | Sorts.QSort _ -> false
 
 let is_impredicative_family env = function
   | Sorts.InSProp | Sorts.InProp -> true
   | Sorts.InSet -> is_impredicative_set env
-  | Sorts.InType -> false
+  | Sorts.InType | Sorts.InQSort -> false
 
 let type_in_type env = not (typing_flags env).check_universes
 let deactivated_guard env = not (typing_flags env).check_guarded
@@ -287,8 +269,6 @@ let set_universes_lbound env lbound = { env with env_universes_lbound = lbound }
 let named_context env = env.env_named_context.env_named_ctx
 let named_context_val env = env.env_named_context
 let rel_context env = env.env_rel_context.env_rel_ctx
-let opaque_tables env = env.indirect_pterms
-let set_opaque_tables env indirect_pterms = { env with indirect_pterms }
 
 let empty_context env =
   match env.env_rel_context.env_rel_ctx, env.env_named_context.env_named_ctx with
@@ -311,7 +291,7 @@ let fold_rel_context f env ~init =
   let rec fold_right env =
     match match_rel_context_val env.env_rel_context with
     | None -> init
-    | Some (rd, _, rc) ->
+    | Some (rd, rc) ->
         let env =
           { env with
             env_rel_context = rc;
@@ -364,7 +344,7 @@ let pop_rel_context n env =
     if Int.equal n 0 then ctx
     else match match_rel_context_val ctx with
     | None -> invalid_arg "List.skipn"
-    | Some (_, _, ctx) -> skip (pred n) ctx
+    | Some (_, ctx) -> skip (pred n) ctx
   in
   let ctxt = env.env_rel_context in
   { env with
@@ -375,7 +355,7 @@ let fold_named_context f env ~init =
   let rec fold_right env =
     match match_named_context_val env.env_named_context with
     | None -> init
-    | Some (d, _v, rem) ->
+    | Some (d, rem) ->
         let env =
           reset_with_named_context rem env in
         f env d (fold_right env)
@@ -397,13 +377,28 @@ let check_constraints c env =
   UGraph.check_constraints c env.env_universes
 
 let add_universes ~lbound ~strict ctx g =
+  let _qs, us = UVars.Instance.to_array (UVars.UContext.instance ctx) in
   let g = Array.fold_left
-            (fun g v -> UGraph.add_universe ~lbound ~strict v g)
-            g (Univ.Instance.to_array (Univ.UContext.instance ctx))
+      (fun g v -> UGraph.add_universe ~lbound ~strict v g)
+      g us
   in
-    UGraph.merge_constraints (Univ.UContext.constraints ctx) g
+  UGraph.merge_constraints (UVars.UContext.constraints ctx) g
+
+let add_qualities qs known =
+  let open Sorts.Quality in
+  Array.fold_left (fun known q ->
+      match q with
+      | QVar q ->
+        let known' = Sorts.QVar.Set.add q known in
+        let () = if known == known' then CErrors.anomaly Pp.(str"multiply bound sort quality") in
+        known'
+      | QConstant _ -> CErrors.anomaly Pp.(str "constant quality in ucontext"))
+    known
+    qs
 
 let push_context ?(strict=false) ctx env =
+  let qs, _us = UVars.Instance.to_array (UVars.UContext.instance ctx) in
+  let env = { env with env_qualities = add_qualities qs env.env_qualities } in
   map_universes (add_universes ~lbound:(universes_lbound env) ~strict ctx) env
 
 let add_universes_set ~lbound ~strict ctx g =
@@ -441,7 +436,6 @@ let same_flags {
      enable_native_compiler;
      impredicative_set;
      sprop_allowed;
-     cumulative_sprop;
      allow_uip;
   } alt =
   check_guarded == alt.check_guarded &&
@@ -454,18 +448,15 @@ let same_flags {
   enable_native_compiler == alt.enable_native_compiler &&
   impredicative_set == alt.impredicative_set &&
   sprop_allowed == alt.sprop_allowed &&
-  cumulative_sprop == alt.cumulative_sprop &&
   allow_uip == alt.allow_uip
 [@warning "+9"]
 
-let set_cumulative_sprop b = map_universes (UGraph.set_cumulative_sprop b)
 let set_type_in_type b = map_universes (UGraph.set_type_in_type b)
 
 let set_typing_flags c env =
   if same_flags env.env_typing_flags c then env
   else
     let env = { env with env_typing_flags = c } in
-    let env = set_cumulative_sprop c.cumulative_sprop env in
     let env = set_type_in_type (not c.check_universes) env in
     env
 
@@ -475,9 +466,6 @@ let update_typing_flags ?typing_flags env =
 let set_impredicative_set b env =
   set_typing_flags {env.env_typing_flags with impredicative_set=b} env
 
-let set_cumulative_sprop b env =
-  set_typing_flags {env.env_typing_flags with cumulative_sprop=b} env
-
 let set_type_in_type b env =
   set_typing_flags {env.env_typing_flags with check_universes=not b} env
 
@@ -497,8 +485,8 @@ let add_constant_key kn cb linkinfo env =
     { env.env_globals with
       Globals.constants = new_constants }
   in
-  let irr_constants = if cb.const_relevance == Sorts.Irrelevant
-    then Cset_env.add kn env.irr_constants
+  let irr_constants = if cb.const_relevance != Sorts.Relevant
+    then Cmap_env.add kn cb.const_relevance env.irr_constants
     else env.irr_constants
   in
   { env with irr_constants; env_globals = new_globals }
@@ -510,20 +498,20 @@ let add_constant kn cb env =
 let constant_type env (kn,u) =
   let cb = lookup_constant kn env in
   let uctx = Declareops.constant_polymorphic_context cb in
-  let csts = Univ.AbstractContext.instantiate u uctx in
+  let csts = UVars.AbstractContext.instantiate u uctx in
   (subst_instance_constr u cb.const_type, csts)
 
 type const_evaluation_result =
   | NoBody
   | Opaque
-  | IsPrimitive of Univ.Instance.t * CPrimitives.t
+  | IsPrimitive of UVars.Instance.t * CPrimitives.t
 
 exception NotEvaluableConst of const_evaluation_result
 
 let constant_value_and_type env (kn, u) =
   let cb = lookup_constant kn env in
   let uctx = Declareops.constant_polymorphic_context cb in
-  let cst = Univ.AbstractContext.instantiate u uctx in
+  let cst = UVars.AbstractContext.instantiate u uctx in
   let b' = match cb.const_body with
     | Def l_body -> Some (subst_instance_constr u l_body)
     | OpaqueDef _ -> None
@@ -597,7 +585,7 @@ let polymorphic_constant cst env =
   Declareops.constant_is_polymorphic (lookup_constant cst env)
 
 let polymorphic_pconstant (cst,u) env =
-  if Univ.Instance.is_empty u then false
+  if UVars.Instance.is_empty u then false
   else polymorphic_constant cst env
 
 let type_in_type_constant cst env =
@@ -611,8 +599,9 @@ let lookup_projection p env =
   match mib.mind_record with
   | NotRecord | FakeRecord -> anomaly ~label:"lookup_projection" Pp.(str "not a projection")
   | PrimRecord infos ->
-    let _,_,_,typs = infos.(i) in
-    typs.(Projection.arg p)
+    let _,_,rs,typs = infos.(i) in
+    let arg = Projection.arg p in
+    rs.(arg), typs.(arg)
 
 let get_projection env ind ~proj_arg =
   let mib = lookup_mind (fst ind) env in
@@ -627,7 +616,7 @@ let polymorphic_ind (mind,_i) env =
   Declareops.inductive_is_polymorphic (lookup_mind mind env)
 
 let polymorphic_pind (ind,u) env =
-  if Univ.Instance.is_empty u then false
+  if UVars.Instance.is_empty u then false
   else polymorphic_ind ind env
 
 let type_in_type_ind (mind,_i) env =
@@ -645,7 +634,7 @@ let template_polymorphic_variables (mind, _) env =
   | None -> []
 
 let template_polymorphic_pind (ind,u) env =
-  if not (Univ.Instance.is_empty u) then false
+  if not (UVars.Instance.is_empty u) then false
   else template_polymorphic_ind ind env
 
 let add_mind_key kn (mind, _ as mind_key) env =
@@ -655,8 +644,8 @@ let add_mind_key kn (mind, _ as mind_key) env =
       Globals.inductives = new_inds; }
   in
   let irr_inds = Array.fold_left_i (fun i irr_inds mip ->
-      if mip.mind_relevance == Sorts.Irrelevant
-      then Indset_env.add (kn, i) irr_inds
+      if mip.mind_relevance != Sorts.Relevant
+      then Indmap_env.add (kn, i) mip.mind_relevance irr_inds
       else irr_inds) env.irr_inds mind.mind_packets
   in
   { env with irr_inds; env_globals = new_globals }
@@ -685,7 +674,7 @@ let constant_context env c =
 let universes_of_global env r =
   let open GlobRef in
     match r with
-    | VarRef _ -> Univ.AbstractContext.empty
+    | VarRef _ -> UVars.AbstractContext.empty
     | ConstRef c -> constant_context env c
     | IndRef (mind,_) | ConstructRef ((mind,_),_) ->
       let mib = lookup_mind mind env in
@@ -781,11 +770,11 @@ let make_judge v tj =
 let j_val j = j.uj_val
 let j_type j = j.uj_type
 
-type 'types punsafe_type_judgment = {
+type ('types, 'sorts) punsafe_type_judgment = {
   utj_val : 'types;
-  utj_type : Sorts.t }
+  utj_type : 'sorts }
 
-type unsafe_type_judgment = types punsafe_type_judgment
+type unsafe_type_judgment = (types, Sorts.t) punsafe_type_judgment
 
 exception Hyp_not_found
 
@@ -793,33 +782,36 @@ let apply_to_hyp ctxt id f =
   let open Context.Named.Declaration in
   let rec aux rtail ctxt =
     match match_named_context_val ctxt with
-    | Some (d, v, ctxt) ->
+    | Some (d, ctxt) ->
         if Id.equal (get_id d) id then
-          push_named_context_val_val (f ctxt.env_named_ctx d rtail) v ctxt
+          push_named_context_val (f ctxt.env_named_ctx d rtail) ctxt
         else
           let ctxt' = aux (d::rtail) ctxt in
-          push_named_context_val_val d v ctxt'
+          push_named_context_val d ctxt'
     | None -> raise Hyp_not_found
   in aux [] ctxt
 
 (* To be used in Logic.clear_hyps *)
-let remove_hyps ids check_context check_value ctxt =
-  let rec remove_hyps ctxt = match match_named_context_val ctxt with
-  | None -> empty_named_context_val, false
-  | Some (d, v, rctxt) ->
-     let open Context.Named.Declaration in
-    let (ans, seen) = remove_hyps rctxt in
-    if Id.Set.mem (get_id d) ids then (ans, true)
-    else if not seen then ctxt, false
-    else
-      let rctxt' = ans in
-      let d' = check_context d in
-      let v' = check_value v in
-      if d == d' && v == v' && rctxt == rctxt' then
-        ctxt, true
-      else push_named_context_val_val d' v' rctxt', true
+let remove_hyps ids check_context ctxt =
+  let rec remove_hyps ids ctxt =
+    if Id.Set.is_empty ids then ctxt, false
+    else match match_named_context_val ctxt with
+    | None -> empty_named_context_val, false
+    | Some (d, rctxt) ->
+      let id0 = Context.Named.Declaration.get_id d in
+      let removed = Id.Set.mem id0 ids in
+      let ids = if removed then Id.Set.remove id0 ids else ids in
+      let (ans, seen) = remove_hyps ids rctxt in
+      if removed then (ans, true)
+      else if not seen then ctxt, false
+      else
+        let rctxt' = ans in
+        let d' = check_context d in
+        if d == d' && rctxt == rctxt' then
+          ctxt, true
+        else push_named_context_val d' rctxt', true
   in
-  fst (remove_hyps ctxt)
+  fst (remove_hyps ids ctxt)
 
 (* A general request *)
 
