@@ -12,6 +12,7 @@ open CErrors
 open Util
 open Names
 open Univ
+open UVars
 open Constr
 open Vars
 open Declarations
@@ -59,7 +60,7 @@ let inductive_nonrec_rec_paramdecls (mib,u) =
   Context.Rel.chop_nhyps nnonrecparamdecls paramdecls
 
 let instantiate_inductive_constraints mib u =
-  Univ.AbstractContext.instantiate u (Declareops.inductive_polymorphic_context mib)
+  UVars.AbstractContext.instantiate u (Declareops.inductive_polymorphic_context mib)
 
 (************************************************************************)
 
@@ -215,9 +216,12 @@ let instantiate_universes ctx (templ, ar) args =
 
 (* Type of an inductive type *)
 
-let relevance_of_inductive env ind =
+let relevance_of_ind_body mip u =
+  UVars.subst_instance_relevance u mip.mind_relevance
+
+let relevance_of_inductive env (ind,u) =
   let _, mip = lookup_mind_specif env ind in
-  mip.mind_relevance
+  relevance_of_ind_body mip u
 
 let check_instance mib u =
   if not (match mib.mind_universes with
@@ -300,15 +304,63 @@ let abstract_constructor_type_relatively_to_inductive_types_context ntyps mind t
 
 (* Get type of inductive, with parameters instantiated *)
 
+(* XXX questionable for sort poly inductives *)
 let inductive_sort_family mip =
   match mip.mind_arity with
   | RegularArity s -> Sorts.family s.mind_sort
   | TemplateArity _ -> Sorts.InType
 
-let mind_arity mip =
-  mip.mind_arity_ctxt, inductive_sort_family mip
+let quality_leq q q' =
+  let open Sorts.Quality in
+  match q, q' with
+  | QVar q, QVar q' -> Sorts.QVar.equal q q'
+  | QConstant q, QConstant q' ->
+    begin match q, q' with
+    | QSProp, _
+    | _, QType
+    | QProp, QProp
+      -> true
+    | (QProp|QType), _ -> false
+    end
+  | (QVar _|QConstant _), _ -> false
 
-let elim_sort (_,mip) = mip.mind_kelim
+type squash = SquashToSet | SquashToQuality of Sorts.Quality.t
+
+let is_squashed ((_,mip),u) =
+  match mip.mind_arity with
+  | TemplateArity _ -> None (* template is never squashed *)
+  | RegularArity a ->
+    match mip.mind_squashed with
+    | None -> None
+    | Some squash ->
+      let indq = Sorts.quality (UVars.subst_instance_sort u a.mind_sort) in
+      match squash with
+      | AlwaysSquashed -> begin match a.mind_sort with
+          | Sorts.Set -> Some SquashToSet
+          | _ -> Some (SquashToQuality indq)
+        end
+      | SometimesSquashed squash ->
+        (* impredicative set squashes are always AlwaysSquashed,
+           so here if inds=Set it is a sort poly squash (see "foo6" in test sort_poly.v) *)
+        if Sorts.Quality.Set.for_all (fun q ->
+            let q = UVars.subst_instance_quality u q in
+            quality_leq q indq)
+            squash
+        then None
+        else Some (SquashToQuality indq)
+
+let is_allowed_elimination specifu s =
+  let open Sorts in
+  match is_squashed specifu with
+  | None -> true
+  | Some SquashToSet ->
+    begin match s with
+      | SProp|Prop|Set -> true
+      | QSort _ | Type _ ->
+        (* XXX in [Type u] case, should we check [u == set] in the ugraph? *)
+        false
+    end
+  | Some (SquashToQuality indq) -> quality_leq (Sorts.quality s) indq
 
 let is_private (mib,_) = mib.mind_private = Some true
 let is_primitive_record (mib,_) =
@@ -347,15 +399,20 @@ let expand_arity (mib, mip) (ind, u) params nas =
   let params = Vars.subst_of_rel_context_instance paramdecl params in
   let realdecls, _ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
   let self =
-    let u = Univ.(Instance.of_array (Array.init (Instance.length u) Level.var)) in
+    let u = Instance.abstract_instance (Instance.length u) in
     let args = Context.Rel.instance mkRel 0 mip.mind_arity_ctxt in
     mkApp (mkIndU (ind, u), args)
   in
-  let na = Context.make_annot Anonymous mip.mind_relevance in
+  let na = Context.make_annot Anonymous (relevance_of_ind_body mip u)  in
   let realdecls = LocalAssum (na, self) :: realdecls in
   instantiate_context u params nas realdecls
 
-let expand_case_specif mib (ci, u, params, p, iv, c, br) =
+type ('constr,'types) pexpanded_case =
+  (case_info * ('constr * Sorts.relevance) * 'constr pcase_invert * 'constr * 'constr array)
+
+type expanded_case = (constr,types) pexpanded_case
+
+let expand_case_specif mib (ci, u, params, (p,rp), iv, c, br) =
   (* Γ ⊢ c : I@{u} params args *)
   (* Γ, indices, self : I@{u} params indices ⊢ p : Type *)
   let mip = mib.mind_packets.(snd ci.ci_ind) in
@@ -376,13 +433,13 @@ let expand_case_specif mib (ci, u, params, p, iv, c, br) =
     in
     Array.map2_i build_one_branch br mip.mind_nf_lc
   in
-  (ci, ep, iv, c, ebr)
+  (ci, (ep,rp), iv, c, ebr)
 
 let expand_case env (ci, _, _, _, _, _, _ as case) =
   let specif = Environ.lookup_mind (fst ci.ci_ind) env in
   expand_case_specif specif case
 
-let contract_case env (ci, p, iv, c, br) =
+let contract_case env (ci, (p,rp), iv, c, br) =
   let (mib, mip) = lookup_mind_specif env ci.ci_ind in
   let (arity, p) = Term.decompose_lambda_n_decls (mip.mind_nrealdecls + 1) p in
   let (u, pms) = match arity with
@@ -400,7 +457,7 @@ let contract_case env (ci, p, iv, c, br) =
   in
   let p =
     let nas = Array.of_list (List.rev_map get_annot arity) in
-    (nas, p)
+    ((nas, p),rp)
   in
   let map i br =
     let (ctx, br) = Term.decompose_lambda_n_decls mip.mind_consnrealdecls.(i) br in
@@ -433,14 +490,13 @@ let build_branches_type (ind,u) (_,mip as specif) params p =
 (************************************************************************)
 (* Checking the case annotation is relevant *)
 
-let check_case_info env (indsp,u) r ci =
+let check_case_info env (indsp,u) ci =
   let (mib,mip as spec) = lookup_mind_specif env indsp in
   if
     not (QInd.equal env indsp ci.ci_ind) ||
     not (Int.equal mib.mind_nparams ci.ci_npar) ||
     not (Array.equal Int.equal mip.mind_consnrealdecls ci.ci_cstr_ndecls) ||
     not (Array.equal Int.equal mip.mind_consnrealargs ci.ci_cstr_nargs) ||
-    not (ci.ci_relevance == r) ||
     is_primitive_record spec
   then raise (TypeError(env,WrongCaseInfo((indsp,u),ci)))
 
@@ -701,9 +757,9 @@ let branches_specif renv c_spec ci =
       car
 
 let check_inductive_codomain env p =
-  let absctx, ar = hnf_decompose_lambda_decls env p in
+  let absctx, ar = whd_decompose_lambda_decls env p in
   let env = push_rel_context absctx env in
-  let arctx, s = hnf_decompose_prod_decls env ar in
+  let arctx, s = whd_decompose_prod_decls env ar in
   let env = push_rel_context arctx env in
   let i,_l' = decompose_app (whd_all env s) in
   isInd i
@@ -717,7 +773,7 @@ let ienv_push_inductive (env, ra_env) ((mind,u),lpar) =
   let mib = Environ.lookup_mind mind env in
   let ntypes = mib.mind_ntypes in
   let push_ind mip env =
-    let r = mip.mind_relevance in
+    let r = relevance_of_ind_body mip u in
     let anon = Context.make_annot Anonymous r in
     let decl = LocalAssum (anon, hnf_prod_applist env (type_of_inductive ((mib,mip),u)) lpar) in
     push_rel decl env
@@ -871,13 +927,13 @@ let restrict_spec env spec p =
   match spec with
   | Not_subterm | Internally_bound_subterm _ -> spec
   | _ ->
-  let absctx, ar = hnf_decompose_lambda_decls env p in
+  let absctx, ar = whd_decompose_lambda_decls env p in
   (* Optimization: if the predicate is not dependent, no restriction is needed
      and we avoid building the recargs tree. *)
   if noccur_with_meta 1 (Context.Rel.length absctx) ar then spec
   else
   let env = push_rel_context absctx env in
-  let arctx, s = hnf_decompose_prod_decls env ar in
+  let arctx, s = whd_decompose_prod_decls env ar in
   let env = push_rel_context arctx env in
   let i,args = decompose_app_list (whd_all env s) in
   match kind i with
@@ -904,7 +960,7 @@ let rec subterm_specif renv stack t =
     match kind f with
     | Rel k -> subterm_var k renv
     | Case (ci, u, pms, p, iv, c, lbr) -> (* iv ignored: it's just a cache *)
-      let (ci, p, _iv, c, lbr) = expand_case renv.env (ci, u, pms, p, iv, c, lbr) in
+      let (ci, (p,_), _iv, c, lbr) = expand_case renv.env (ci, u, pms, p, iv, c, lbr) in
        let stack' = push_stack_closures renv l stack in
        let cases_spec =
          branches_specif renv (lazy_subterm_specif renv [] c) ci
@@ -925,7 +981,7 @@ let rec subterm_specif renv stack t =
       *)
     if not (check_inductive_codomain renv.env typarray.(i)) then Not_subterm
     else
-      let (ctxt,clfix) = hnf_decompose_prod renv.env typarray.(i) in
+      let (ctxt,clfix) = whd_decompose_prod renv.env typarray.(i) in
       let oind =
         let env' = push_rel_context ctxt renv.env in
           try Some(fst(find_inductive env' clfix))
@@ -945,7 +1001,7 @@ let rec subterm_specif renv stack t =
         let decrArg = recindxs.(i) in
         let theBody = bodies.(i)   in
         let nbOfAbst = decrArg+1 in
-        let sign,strippedBody = hnf_decompose_lambda_n_decls renv.env nbOfAbst theBody in
+        let sign,strippedBody = whd_decompose_lambda_n_assum renv.env nbOfAbst theBody in
                    (* pushing the fix parameters *)
         let stack' = push_stack_closures renv l stack in
         let renv'' = push_ctxt_renv renv' sign in
@@ -965,7 +1021,7 @@ let rec subterm_specif renv stack t =
       (* Metas and evars are considered OK *)
     | (Meta _|Evar _) -> Dead_code
 
-    | Proj (p, c) ->
+    | Proj (p, _, c) ->
       let subt = subterm_specif renv stack c in
       (match subt with
        | Subterm (_, _s, wf) ->
@@ -1089,7 +1145,7 @@ let filter_stack_domain env nr p stack =
     match kind t with
     | Prod (n,a,c0) ->
       let d = LocalAssum (n,a) in
-      let ctx, a = hnf_decompose_prod_decls env a in
+      let ctx, a = whd_decompose_prod_decls env a in
       let env = push_rel_context ctx env in
       let ty, args = decompose_app_list (whd_all env a) in
       let elt = match kind ty with
@@ -1163,7 +1219,7 @@ let check_one_fix renv recpos trees def =
             end
 
         | Case (ci, u, pms, ret, iv, c_0, br) -> (* iv ignored: it's just a cache *)
-            let (ci, p, _iv, c_0, brs) = expand_case renv.env (ci, u, pms, ret, iv, c_0, br) in
+            let (ci, (p,_), _iv, c_0, brs) = expand_case renv.env (ci, u, pms, ret, iv, c_0, br) in
             let needreduce_c_0, rs = check_rec_call renv rs c_0 in
             let rs = check_inert_subterm_rec_call renv rs p in
             (* compute the recarg info for the arguments of each branch *)
@@ -1285,7 +1341,7 @@ let check_one_fix renv recpos trees def =
         | Ind _ | Construct _ ->
             check_rec_call_state renv NoNeedReduce stack rs (fun () -> None)
 
-        | Proj (p, c) ->
+        | Proj (p, _, c) ->
             begin
               let needreduce', rs = check_rec_call renv rs c in
               check_rec_call_state renv needreduce' stack rs (fun () ->
@@ -1425,16 +1481,17 @@ let inductive_of_mutfix env ((nvect,bodynum),(names,types,bodies as recdef)) =
             else anomaly ~label:"check_one_fix" (Pp.str "Bad occurrence of recursive call.")
         | _ -> raise_err env i NotEnoughAbstractionInFixBody
     in
-    let ((ind, _), _) as res = check_occur fixenv 1 def in
+    let ((ind, u), _) as res = check_occur fixenv 1 def in
     let _, mip = lookup_mind_specif env ind in
     (* recursive sprop means non record with projections -> squashed *)
-    if mip.mind_relevance == Sorts.Irrelevant &&
-       not (Environ.is_type_in_type env (GlobRef.IndRef ind))
-    then
-      begin
-        if names.(i).Context.binder_relevance == Sorts.Relevant
-        then raise_err env i FixpointOnIrrelevantInductive
-      end;
+    let () =
+      if Environ.is_type_in_type env (GlobRef.IndRef ind) then ()
+      else match relevance_of_ind_body mip u with
+        | Sorts.Irrelevant | Sorts.RelevanceVar _ as rind ->
+          if not (Sorts.relevance_equal names.(i).Context.binder_relevance rind)
+          then raise_err env i FixpointOnIrrelevantInductive
+        | Sorts.Relevant -> ()
+    in
     res
   in
   (* Do it on every fixpoint *)
@@ -1535,7 +1592,7 @@ let check_one_cofix env nbfix def deftype =
 
         | Case (ci, u, pms, p, iv, tm, br) -> (* iv ignored: just a cache *)
           begin
-            let (_, p, _iv, tm, vrest) = expand_case env (ci, u, pms, p, iv, tm, br) in
+            let (_, (p,_), _iv, tm, vrest) = expand_case env (ci, u, pms, p, iv, tm, br) in
             let tree = match restrict_spec env (Subterm (Int.Set.empty, Strict, tree)) p with
             | Dead_code -> assert false
             | Subterm (_, _, tree') -> tree'
