@@ -19,8 +19,6 @@ open Genlambda
 open Nativelambda
 open Environ
 
-[@@@ocaml.warning "-37"]
-
 (** This file defines the mllambda code generation phase of the native
 compiler. mllambda represents a fragment of ML, and can easily be printed
 to OCaml code. *)
@@ -173,9 +171,20 @@ let eq_symbol sy1 sy2 =
   | SymbMatch sw1, SymbMatch sw2 -> eq_annot_sw sw1 sw2
   | SymbInd ind1, SymbInd ind2 -> Ind.CanOrd.equal ind1 ind2
   | SymbEvar evk1, SymbEvar evk2 -> Evar.equal evk1 evk2
-  | SymbLevel l1, SymbLevel l2 -> Univ.Level.equal l1 l2
+  | SymbInstance u1, SymbInstance u2 -> UVars.Instance.equal u1 u2
   | SymbProj (i1, k1), SymbProj (i2, k2) -> Ind.CanOrd.equal i1 i2 && Int.equal k1 k2
-  | _, _ -> false
+
+  | (SymbValue _
+    | SymbSort _
+    | SymbName _
+    | SymbConst _
+    | SymbMatch _
+    | SymbInd _
+    | SymbEvar _
+    | SymbInstance _
+    | SymbProj _), _
+    -> false
+
 
 let hash_symbol symb =
   match symb with
@@ -186,7 +195,7 @@ let hash_symbol symb =
   | SymbMatch sw -> combinesmall 5 (hash_annot_sw sw)
   | SymbInd ind -> combinesmall 6 (Ind.CanOrd.hash ind)
   | SymbEvar evk -> combinesmall 7 (Evar.hash evk)
-  | SymbLevel l -> combinesmall 8 (Univ.Level.hash l)
+  | SymbInstance u -> combinesmall 8 (UVars.Instance.hash u)
   | SymbProj (i, k) -> combinesmall 9 (combine (Ind.CanOrd.hash i) k)
 
 module HashedTypeSymbol = struct
@@ -236,10 +245,10 @@ let get_evar tbl i =
     | SymbEvar ev -> ev
     | _ -> anomaly (Pp.str "get_evar failed.")
 
-let get_level tbl i =
+let get_instance tbl i =
   match tbl.(i) with
-    | SymbLevel u -> u
-    | _ -> anomaly (Pp.str "get_level failed.")
+    | SymbInstance u -> u
+    | _ -> anomaly (Pp.str "get_instance failed.")
 
 let get_proj tbl i =
   match tbl.(i) with
@@ -280,25 +289,13 @@ type primitive =
   | Mk_uint
   | Mk_float
   | Mk_int
-  | Mk_bool
   | Val_to_int
   | Mk_evar
   | MLand
   | MLnot
-  | MLle
-  | MLlt
-  | MLinteq
-  | MLlsl
-  | MLlsr
   | MLland
-  | MLlor
-  | MLlxor
-  | MLadd
-  | MLsub
-  | MLmul
   | MLmagic
-  | MLarrayget
-  | Mk_empty_instance
+  | MLsubst_instance_instance
   | MLparray_of_array
   | Get_value
   | Get_sort
@@ -307,31 +304,100 @@ type primitive =
   | Get_match
   | Get_ind
   | Get_evar
-  | Get_level
+  | Get_instance
   | Get_proj
   | Get_symbols
   | Lazy
   | Coq_primitive of CPrimitives.t * bool (* check for accu *)
+  | Mk_empty_instance
 
 let eq_primitive p1 p2 =
   match p1, p2 with
-  | Mk_prod, Mk_prod -> true
-  | Mk_sort, Mk_sort -> true
-  | Mk_ind, Mk_ind -> true
-  | Mk_const, Mk_const -> true
-  | Mk_sw, Mk_sw -> true
+  | Mk_prod, Mk_prod
+  | Mk_sort, Mk_sort
+  | Mk_ind, Mk_ind
+  | Mk_const, Mk_const
+  | Mk_sw, Mk_sw
+  | Mk_proj, Mk_proj
+  | Is_int, Is_int
+  | Is_float, Is_float
+  | Is_parray, Is_parray
+  | Cast_accu, Cast_accu
+  | Upd_cofix, Upd_cofix
+  | Force_cofix, Force_cofix
+  | Mk_uint, Mk_uint
+  | Mk_float, Mk_float
+  | Mk_int, Mk_int
+  | Val_to_int, Val_to_int
+  | Mk_evar, Mk_evar
+  | MLand, MLand
+  | MLnot, MLnot
+  | MLland, MLland
+  | MLmagic, MLmagic
+  | MLsubst_instance_instance, MLsubst_instance_instance
+  | MLparray_of_array, MLparray_of_array
+  | Get_value, Get_value
+  | Get_sort, Get_sort
+  | Get_name, Get_name
+  | Get_const, Get_const
+  | Get_match, Get_match
+  | Get_ind, Get_ind
+  | Get_evar, Get_evar
+  | Get_instance, Get_instance
+  | Get_proj, Get_proj
+  | Get_symbols, Get_symbols
+  | Lazy, Lazy
+  | Mk_empty_instance, Mk_empty_instance
+    -> true
+
   | Mk_fix (rp1, i1), Mk_fix (rp2, i2) -> Int.equal i1 i2 && eq_rec_pos rp1 rp2
   | Mk_cofix i1, Mk_cofix i2 -> Int.equal i1 i2
   | Mk_rel i1, Mk_rel i2 -> Int.equal i1 i2
   | Mk_var id1, Mk_var id2 -> Id.equal id1 id2
-  | Cast_accu, Cast_accu -> true
-  | Upd_cofix, Upd_cofix -> true
-  | Force_cofix, Force_cofix -> true
-  | Mk_evar, Mk_evar -> true
-  | Mk_proj, Mk_proj -> true
-  | MLarrayget, MLarrayget -> true
+  | Coq_primitive (prim1,b1), Coq_primitive (prim2,b2) ->
+    CPrimitives.equal prim1 prim2 && Bool.equal b1 b2
 
-  | _ -> false
+  | (Mk_prod
+    | Mk_sort
+    | Mk_ind
+    | Mk_const
+    | Mk_sw
+    | Mk_fix _
+    | Mk_cofix _
+    | Mk_rel _
+    | Mk_var _
+    | Mk_proj
+    | Is_int
+    | Is_float
+    | Is_parray
+    | Cast_accu
+    | Upd_cofix
+    | Force_cofix
+    | Mk_uint
+    | Mk_float
+    | Mk_int
+    | Val_to_int
+    | Mk_evar
+    | MLand
+    | MLnot
+    | MLland
+    | MLmagic
+    | MLsubst_instance_instance
+    | MLparray_of_array
+    | Get_value
+    | Get_sort
+    | Get_name
+    | Get_const
+    | Get_match
+    | Get_ind
+    | Get_evar
+    | Get_instance
+    | Get_proj
+    | Get_symbols
+    | Lazy
+    | Coq_primitive _
+    | Mk_empty_instance), _
+    -> false
 
 let primitive_hash = function
   | Mk_prod -> 1
@@ -354,42 +420,31 @@ let primitive_hash = function
   | Force_cofix -> 14
   | Mk_uint -> 15
   | Mk_int -> 16
-  | Mk_bool -> 17
-  | Val_to_int -> 18
-  | Mk_evar -> 19
-  | MLand -> 20
-  | MLle -> 21
-  | MLlt -> 22
-  | MLinteq -> 23
-  | MLlsl -> 24
-  | MLlsr -> 25
-  | MLland -> 26
-  | MLlor -> 27
-  | MLlxor -> 28
-  | MLadd -> 29
-  | MLsub -> 30
-  | MLmul -> 31
-  | MLmagic -> 32
-  | Coq_primitive (prim, b) -> combinesmall 33 (combine (CPrimitives.hash prim) (Hashtbl.hash b))
-  | Mk_proj -> 34
-  | MLarrayget -> 35
-  | Mk_empty_instance -> 36
-  | Mk_float -> 37
-  | Is_float -> 38
-  | Is_parray -> 39
-  | MLnot -> 40
-  | MLparray_of_array -> 41
-  | Get_value -> 42
-  | Get_sort -> 43
-  | Get_name -> 44
-  | Get_const -> 45
-  | Get_match -> 46
-  | Get_ind -> 47
-  | Get_evar -> 48
-  | Get_level -> 49
-  | Get_proj -> 50
-  | Get_symbols -> 51
-  | Lazy -> 52
+  | Val_to_int -> 17
+  | Mk_evar -> 18
+  | MLand -> 19
+  | MLland -> 20
+  | MLmagic -> 21
+  | Coq_primitive (prim, b) -> combinesmall 22 (combine (CPrimitives.hash prim) (Hashtbl.hash b))
+  | Mk_proj -> 23
+  | MLsubst_instance_instance -> 24
+  | Mk_float -> 25
+  | Is_float -> 26
+  | Is_parray -> 27
+  | MLnot -> 28
+  | MLparray_of_array -> 29
+  | Get_value -> 30
+  | Get_sort -> 31
+  | Get_name -> 32
+  | Get_const -> 33
+  | Get_match -> 34
+  | Get_ind -> 35
+  | Get_evar -> 36
+  | Get_instance -> 37
+  | Get_proj -> 38
+  | Get_symbols -> 39
+  | Lazy -> 40
+  | Mk_empty_instance -> 41
 
 type mllambda =
   | MLlocal        of lname
@@ -959,8 +1014,8 @@ let get_evar_code i =
   MLprimitive (Get_evar,
     [|MLglobal symbols_tbl_name; MLint i|])
 
-let get_level_code i =
-  MLprimitive (Get_level,
+let get_instance_code i =
+  MLprimitive (Get_instance,
     [|MLglobal symbols_tbl_name; MLint i|])
 
 let get_proj_code i =
@@ -1008,6 +1063,8 @@ let merge_branches t =
 
 let app_prim p args = MLprimitive (p, args)
 
+let ml_empty_instance = MLprimitive (Mk_empty_instance, [||])
+
 type prim_aux =
   | PAprim of string * pconstant * CPrimitives.t * prim_aux array
   | PAml of mllambda
@@ -1055,19 +1112,23 @@ let cast_to_int v =
   | _ -> MLprimitive (Val_to_int, [|v|])
 
 let ml_of_instance instance u =
-  let ml_of_level l =
-    match Univ.Level.var_index l with
-    | Some i ->
-      (* FIXME: use a proper cast function *)
-       let univ = MLprimitive (MLmagic, [|MLlocal (Option.get instance)|]) in
-       MLprimitive (MLarrayget, [|univ; MLint i|])
-    | None -> let i = push_symbol (SymbLevel l) in get_level_code i
-  in
-  let u = Univ.Instance.to_array u in
-  if Array.is_empty u then [||]
-  else let u = Array.map ml_of_level u in
-      (* FIXME: use a proper cast function *)
-       [|MLprimitive (MLmagic, [|MLarray u|])|]
+  if UVars.Instance.is_empty u then [||]
+  else
+    let i = push_symbol (SymbInstance u) in
+    let u_code = get_instance_code i in
+    let has_variable =
+      let qs, us = UVars.Instance.to_array u in
+      Array.exists (fun q -> Option.has_some (Sorts.Quality.var_index q)) qs
+      || Array.exists (fun u -> Option.has_some (Univ.Level.var_index u)) us
+    in
+    let u_code =
+      if has_variable then
+        (* if there are variables then [instance] guaranteed non-None *)
+        let univ = MLprimitive (MLmagic, [|MLlocal (Option.get instance)|]) in
+        MLprimitive (MLsubst_instance_instance, [|univ; u_code|])
+      else u_code
+    in
+    [|MLprimitive (MLmagic, [|u_code|])|]
 
 let compile_prim env decl cond paux =
 
@@ -1187,7 +1248,6 @@ let compile_prim env decl cond paux =
       let annot =
         let (ci, tbl, knd) = annot in {
           asw_ind = ci.ci_ind;
-          asw_ci = ci;
           asw_reloc = tbl;
           asw_finite = knd <> CoFinite;
           asw_prefix = env.env_mind_prefix (fst ci.ci_ind);
@@ -1403,7 +1463,7 @@ let compile_prim env decl cond paux =
   | Lsort s ->
     let i = push_symbol (SymbSort s) in
     let uarg = match env.env_univ with
-      | None -> MLarray [||]
+      | None -> ml_empty_instance
       | Some u -> MLlocal u
     in
     (* FIXME: use a dedicated cast function *)
@@ -1851,6 +1911,7 @@ let pp_mllam fmt l =
     | Mk_var id ->
         Format.fprintf fmt "mk_var_accu (Names.Id.of_string \"%s\")" (string_of_id id)
     | Mk_proj -> Format.fprintf fmt "mk_proj_accu"
+    | Mk_empty_instance -> Format.fprintf fmt "UVars.Instance.empty"
     | Is_int -> Format.fprintf fmt "is_int"
     | Is_float -> Format.fprintf fmt "is_float"
     | Is_parray -> Format.fprintf fmt "is_parray"
@@ -1860,25 +1921,13 @@ let pp_mllam fmt l =
     | Mk_uint -> Format.fprintf fmt "mk_uint"
     | Mk_float -> Format.fprintf fmt "mk_float"
     | Mk_int -> Format.fprintf fmt "mk_int"
-    | Mk_bool -> Format.fprintf fmt "mk_bool"
     | Val_to_int -> Format.fprintf fmt "val_to_int"
     | Mk_evar -> Format.fprintf fmt "mk_evar_accu"
     | MLand -> Format.fprintf fmt "(&&)"
     | MLnot -> Format.fprintf fmt "not"
-    | MLle -> Format.fprintf fmt "(<=)"
-    | MLlt -> Format.fprintf fmt "(<)"
-    | MLinteq -> Format.fprintf fmt "(==)"
-    | MLlsl -> Format.fprintf fmt "(lsl)"
-    | MLlsr -> Format.fprintf fmt "(lsr)"
     | MLland -> Format.fprintf fmt "(land)"
-    | MLlor -> Format.fprintf fmt "(lor)"
-    | MLlxor -> Format.fprintf fmt "(lxor)"
-    | MLadd -> Format.fprintf fmt "(+)"
-    | MLsub -> Format.fprintf fmt "(-)"
-    | MLmul -> Format.fprintf fmt "( * )"
     | MLmagic -> Format.fprintf fmt "Obj.magic"
-    | MLarrayget -> Format.fprintf fmt "Array.get"
-    | Mk_empty_instance -> Format.fprintf fmt "Univ.Instance.empty"
+    | MLsubst_instance_instance -> Format.fprintf fmt "UVars.subst_instance_instance"
     | MLparray_of_array -> Format.fprintf fmt "parray_of_array"
     | Coq_primitive (op, false) ->
        Format.fprintf fmt "no_check_%s" (CPrimitives.to_string op)
@@ -1890,7 +1939,7 @@ let pp_mllam fmt l =
     | Get_match -> Format.fprintf fmt "get_match"
     | Get_ind -> Format.fprintf fmt "get_ind"
     | Get_evar -> Format.fprintf fmt "get_evar"
-    | Get_level -> Format.fprintf fmt "get_level"
+    | Get_instance -> Format.fprintf fmt "get_instance"
     | Get_proj -> Format.fprintf fmt "get_proj"
     | Get_symbols -> Format.fprintf fmt "get_symbols"
     | Lazy -> Format.fprintf fmt "lazy"
@@ -1912,7 +1961,7 @@ let type_of_global gn c = match gn with
   | _ -> match c with
     | MLprimitive (Lazy, _) -> " : Nativevalues.t Lazy.t"
     | MLlam ([|_|], MLprimitive (Lazy, _)) -> " : Nativevalues.t -> Nativevalues.t Lazy.t"
-    | MLprimitive ((Mk_ind | Mk_const), [|_|]) -> " : Univ.Level.t array -> Nativevalues.t"
+    | MLprimitive ((Mk_ind | Mk_const), [|_|]) -> " : UVars.Instance.t -> Nativevalues.t"
     | MLsetref (_,_) -> " : unit"
     | _ -> " : Nativevalues.t"
 
@@ -2013,7 +2062,7 @@ and compile_named env sigma univ auxdefs id =
       Glet(Gnamed id, MLprimitive (Mk_var id, [||]))::auxdefs
 
 let compile_constant env sigma con cb =
-    let no_univs = 0 = Univ.AbstractContext.size (Declareops.constant_polymorphic_context cb) in
+    let no_univs = (0,0) = UVars.AbstractContext.size (Declareops.constant_polymorphic_context cb) in
     begin match cb.const_body with
     | Def t ->
       let code = lambda_of_constr env sigma t in
@@ -2038,7 +2087,7 @@ let compile_constant env sigma con cb =
     | _ ->
         let i = push_symbol (SymbConst con) in
         let args =
-          if no_univs then [|get_const_code i; MLarray [||]|]
+          if no_univs then [|get_const_code i; ml_empty_instance|]
           else [|get_const_code i|]
         in
         (*
@@ -2074,22 +2123,17 @@ let compile_mind mb mind stack =
     let name = Gind ("", ind) in
     let accu =
       let args =
-        if Int.equal (Univ.AbstractContext.size u) 0 then
-          [|get_ind_code j; MLarray [||]|]
+        if (UVars.AbstractContext.size u) = (0,0) then
+          [|get_ind_code j; ml_empty_instance|]
         else [|get_ind_code j|]
       in
       (* FIXME: pass universes here *)
       Glet(name, MLprimitive (Mk_ind, args))
     in
-    let nparams = mb.mind_nparams in
     let add_proj proj_arg acc _pb =
       let tbl = ob.mind_reloc_tbl in
       (* Building info *)
-      let ci = { ci_ind = ind; ci_npar = nparams;
-                 ci_cstr_nargs = [|0|]; ci_relevance = ob.mind_relevance;
-                 ci_cstr_ndecls = [||] (*FIXME*);
-                 ci_pp_info = { ind_tags = []; cstr_tags = [||] (*FIXME*); style = RegularStyle } } in
-      let asw = { asw_ind = ind; asw_prefix = ""; asw_ci = ci;
+      let asw = { asw_ind = ind; asw_prefix = "";
                   asw_reloc = tbl; asw_finite = true } in
       let c_uid = fresh_lname Anonymous in
       let cf_uid = fresh_lname Anonymous in
@@ -2175,7 +2219,7 @@ let compile_deps env sigma prefix init t =
       let const_updates = Cmap_env.add c upd const_updates in
       comp_stack, (mind_updates, const_updates)
   | Construct (((mind,_),_),_u) -> compile_mind_deps env prefix init mind
-  | Proj (p,c) ->
+  | Proj (p,_,c) ->
     let init = compile_mind_deps env prefix init (Projection.mind p) in
     aux env lvl init c
   | Case (ci, _u, _pms, _p, _iv, _c, _ac) ->
@@ -2203,7 +2247,7 @@ let compile_deps env sigma prefix init t =
   aux env 0 init t
 
 let compile_constant_field env con acc cb =
-    let gl = compile_constant env empty_evars con cb in
+    let gl = compile_constant env (empty_evars env) con cb in
     gl@acc
 
 let compile_mind_field mp l acc mb =
