@@ -503,7 +503,7 @@ let safe_push_named d env =
 
 let push_named_def (id,de) senv =
   let sections = get_section senv.sections in
-  let c, r, typ = Constant_typing.translate_local_def senv.env id de in
+  let c, r, typ = Constant_typing.infer_local_def senv.env id de in
   let d = LocalDef (Context.make_annot id r, c, typ) in
   let env'' = safe_push_named d senv.env in
   let sections = Section.push_local d sections in
@@ -511,7 +511,7 @@ let push_named_def (id,de) senv =
 
 let push_named_assum (x,t) senv =
   let sections = get_section senv.sections in
-  let t, r = Constant_typing.translate_local_assum senv.env t in
+  let t, r = Constant_typing.infer_local_assum senv.env t in
   let d = LocalAssum (Context.make_annot x r, t) in
   let sections = Section.push_local d sections in
   let env'' = safe_push_named d senv.env in
@@ -521,12 +521,11 @@ let push_section_context uctx senv =
   let sections = get_section senv.sections in
   let sections = Section.push_local_universe_context uctx sections in
   let senv = { senv with sections=Some sections } in
-  let ctx = Univ.ContextSet.of_context uctx in
-  (* We check that the universes are fresh. FIXME: This should be done
-     implicitly, but we have to work around the API. *)
-  let () = assert (Univ.Level.Set.for_all (fun u -> not (Univ.Level.Set.mem u (fst senv.univ))) (fst ctx)) in
+  let qualities, ctx = UVars.UContext.to_context_set uctx in
+  assert (Sorts.Quality.Set.is_empty qualities);
+  (* push_context checks freshness *)
   { senv with
-    env = Environ.push_context_set ~strict:false ctx senv.env;
+    env = Environ.push_context ~strict:false uctx senv.env;
     univ = Univ.ContextSet.union ctx senv.univ }
 
 (** {6 Insertion of new declarations to current environment } *)
@@ -606,7 +605,7 @@ type global_declaration =
 type exported_opaque = {
   exp_handle : Opaqueproof.opaque_handle;
   exp_body : Constr.t;
-  exp_univs : int option;
+  exp_univs : (int * int) option;
   (* Minimal amount of data needed to rebuild the private universes. We enforce
      in the API that private constants have no internal constraints. *)
 }
@@ -739,7 +738,7 @@ let constant_entry_of_side_effect eff =
     | Monomorphic ->
       Monomorphic_entry
     | Polymorphic auctx ->
-      Polymorphic_entry (Univ.AbstractContext.repr auctx)
+      Polymorphic_entry (UVars.AbstractContext.repr auctx)
   in
   let p =
     match cb.const_body with
@@ -770,8 +769,8 @@ let is_empty_private = function
 
 (* Special function to call when the body of an opaque definition is provided.
   It performs the type-checking of the body immediately. *)
-let translate_direct_opaque ~sec_univs env kn ce =
-  let cb, ctx = Constant_typing.translate_opaque ~sec_univs env kn ce in
+let infer_direct_opaque ~sec_univs env ce =
+  let cb, ctx = Constant_typing.infer_opaque ~sec_univs env ce in
   let body = ce.Entries.opaque_entry_body, Univ.ContextSet.empty in
   let handle _env c () = (c, Univ.ContextSet.empty, 0) in
   let (c, u) = Constant_typing.check_delayed handle ctx (body, ()) in
@@ -804,14 +803,13 @@ let export_side_effects senv eff =
         let env = Environ.push_context_set ~strict:true uctx env in
         let univs = Univ.ContextSet.union uctx univs in
         let env, cb =
-          let kn = eff.seff_constant in
           let ce = constant_entry_of_side_effect eff in
           let open Entries in
           let cb = match ce with
             | DefinitionEff ce ->
-              Constant_typing.translate_constant ~sec_univs env kn (DefinitionEntry ce)
+              Constant_typing.infer_constant ~sec_univs env (DefinitionEntry ce)
             | OpaqueEff ce ->
-              translate_direct_opaque ~sec_univs env kn ce
+              infer_direct_opaque ~sec_univs env ce
           in
           let eff = { eff with seff_body = cb } in
           (push_seff env eff, export_eff eff)
@@ -830,12 +828,12 @@ let export_private_constants eff senv =
   let senv = push_context_set ~strict:true uctx senv in
   let map senv (kn, c) = match c.const_body with
   | OpaqueDef body ->
-    (* Don't care about the body, it has been checked by {!translate_direct_opaque} *)
+    (* Don't care about the body, it has been checked by {!infer_direct_opaque} *)
     let senv, o = push_opaque_proof senv in
     let (_, _, _, h) = Opaqueproof.repr o in
     let univs = match c.const_universes with
     | Monomorphic -> None
-    | Polymorphic auctx -> Some (Univ.AbstractContext.size auctx)
+    | Polymorphic auctx -> Some (UVars.AbstractContext.size auctx)
     in
     let body = Constr.hcons body in
     let opaque = { exp_body = body; exp_handle = h; exp_univs = univs } in
@@ -857,7 +855,7 @@ let add_constant l decl senv =
       match decl with
       | OpaqueEntry ce ->
         let senv, o = push_opaque_proof senv in
-        let cb, ctx = Constant_typing.translate_opaque ~sec_univs senv.env kn ce in
+        let cb, ctx = Constant_typing.infer_opaque ~sec_univs senv.env ce in
         (* Push the delayed data in the environment *)
         let (_, _, _, i) = Opaqueproof.repr o in
         let nonce = Nonce.create () in
@@ -865,7 +863,7 @@ let add_constant l decl senv =
         let senv = { senv with future_cst } in
         senv, { cb with const_body = OpaqueDef o }
       | ConstantEntry ce ->
-        senv, Constant_typing.translate_constant ~sec_univs senv.env kn ce
+        senv, Constant_typing.infer_constant ~sec_univs senv.env ce
   in
   let senv = add_constant_aux senv (kn, cb) in
 
@@ -950,10 +948,10 @@ let add_private_constant l uctx decl senv : (Constant.t * private_constants) * s
       match decl with
       | OpaqueEff ce ->
         let () = assert (check_constraints uctx ce.Entries.opaque_entry_universes) in
-        translate_direct_opaque ~sec_univs senv.env kn ce
+        infer_direct_opaque ~sec_univs senv.env ce
       | DefinitionEff ce ->
         let () = assert (check_constraints uctx ce.Entries.const_entry_universes) in
-        Constant_typing.translate_constant ~sec_univs senv.env kn (Entries.DefinitionEntry ce)
+        Constant_typing.infer_constant ~sec_univs senv.env (Entries.DefinitionEntry ce)
     in
   let dcb = match cb.const_body with
   | Def _ as const_body -> { cb with const_body }
@@ -1293,6 +1291,8 @@ let start_library dir senv =
 
 let export ~output_native_objects senv dir =
   let () = check_current_library dir senv in
+  (* qualities are in the senv only during sections *)
+  let () = assert (Sorts.QVar.Set.is_empty senv.env.Environ.env_qualities) in
   let mp = senv.modpath in
   let str = NoFunctor (List.rev senv.revstruct) in
   let mb =
