@@ -16,7 +16,7 @@ open Environ
 
 (** Should only be used for monomorphic terms *)
 val compile :
-  fail_on_error:bool -> ?universes:int ->
+  fail_on_error:bool -> ?universes:int*int ->
   env -> Genlambda.evars -> constr ->
   (to_patch * fv) option
 (** init, fun, fv *)
