@@ -22,6 +22,11 @@ val (//) : unix_path -> string -> unix_path
 
 val exists_dir : unix_path -> bool
 
+(** [mkdir path] ensures that [path] exists as a directory, creating
+    the missing suffix if necessary (like Unix' mkdirhier) *)
+
+val mkdir : unix_path -> unit
+
 (** [exclude_search_in_dirname path] excludes [path] when processing
     directories *)
 
@@ -41,6 +46,8 @@ val process_subdirectories : (unix_path -> string -> unit) -> unix_path -> unit
 
 (** {6 Files and load paths} *)
 
+val warn_cannot_open_dir : ?loc:Loc.t -> string -> unit
+
 (** Load path entries remember the original root
     given by the user. For efficiency, we keep the full path (field
     [directory]), the root path and the path relative to the root. *)
@@ -124,6 +131,20 @@ type 'a transaction_result = (('a * duration), (Exninfo.iexn * duration)) Result
 val measure_duration : ('a -> 'b) -> 'a -> 'b transaction_result
 val fmt_transaction_result : 'a transaction_result -> Pp.t
 
+(** {6 Instruction count.} *)
+
+type instruction_count = (Int64.t, string) Result.t
+
+val instructions_between : c_start:instruction_count -> c_end:instruction_count -> instruction_count
+val instruction_count_add : instruction_count -> instruction_count -> instruction_count
+
+type 'a instructions_result =
+  (('a * instruction_count), (Exninfo.iexn * instruction_count)) Result.t
+
+val count_instructions : ('a -> 'b) -> 'a -> 'b instructions_result
+
+val fmt_instructions_result : 'a instructions_result -> Pp.t
+
 (** [get_toplevel_path program] builds a complete path to the
    executable denoted by [program]. This involves:
 
