@@ -38,10 +38,10 @@ let _False      = lazy (Coqlib.lib_ref "core.False.type")
 let _not        = lazy (Coqlib.lib_ref "core.not.type")
 
 let whd env sigma t =
-  Reductionops.clos_whd_flags CClosure.betaiotazeta env sigma t
+  Reductionops.clos_whd_flags RedFlags.betaiotazeta env sigma t
 
 let whd_delta env sigma t =
-  Reductionops.clos_whd_flags CClosure.all env sigma t
+  Reductionops.clos_whd_flags RedFlags.all env sigma t
 
 let whd_in_concl =
   reduct_in_concl ~cast:true ~check:false (Reductionops.whd_all, DEFAULTcast)
@@ -83,7 +83,7 @@ let rec decompose_term env sigma t =
         let u = EInstance.kind sigma u in
         let canon_const = Constant.make1 (Constant.canonical c) in
           ATerm.mkSymb (Constr.mkConstU (canon_const,u))
-    | Proj (p, c) ->
+    | Proj (p, _, c) ->
         let canon_mind kn = MutInd.make1 (MutInd.canonical kn) in
         let p' = Projection.map canon_mind p in
         let c = Retyping.expand_projection env sigma p' c [] in
@@ -440,7 +440,7 @@ let cc_tactic depth additional_terms b =
     match sol with
       None -> Tacticals.tclFAIL (str (if b then "simple congruence failed" else "congruence failed"))
     | Some reason ->
-      debug_congruence (fun () -> Pp.str "Goal solved, generating proof ...");
+      Proofview.tclORELSE (debug_congruence (fun () -> Pp.str "Goal solved, generating proof ...");
       match reason with
         Discrimination (i,ipac,j,jpac) ->
         let p=build_proof (Tacmach.pf_env gl) sigma uf (`Discr (i,ipac,j,jpac)) in
@@ -449,7 +449,7 @@ let cc_tactic depth additional_terms b =
       | Incomplete terms_to_complete ->
         let open Glob_term in
         let env = Proofview.Goal.env gl in
-        let hole = DAst.make @@ GHole (Evar_kinds.InternalHole, Namegen.IntroAnonymous) in
+        let hole = DAst.make @@ GHole (GInternalHole) in
         let pr_missing (c, missing) =
           let c = Detyping.detype Detyping.Now Id.Set.empty env sigma c in
           let holes = List.init missing (fun _ -> hole) in
@@ -483,7 +483,14 @@ let cc_tactic depth additional_terms b =
         | HeqG id ->
           convert_to_goal_tac id ta tb p
         | HeqnH (ida,idb) ->
-          convert_to_hyp_tac ida ta idb tb p
+          convert_to_hyp_tac ida ta idb tb p)
+      begin function (e, info) -> match e with
+        | Tactics.NotConvertible ->
+          Tacticals.tclFAIL
+            (str (if b then "simple congruence failed" else "congruence failed") ++
+             str " (cannot build a well-typed proof)")
+        | e -> Proofview.tclZERO ~info e
+      end
   end
 
 let id t = mkLambda (make_annot Anonymous Sorts.Relevant, t, mkRel 1)
