@@ -335,7 +335,7 @@ let string_of_genarg_arg (ArgumentType arg) =
   let pr_targ prtac symb arg = match symb with
   | Extend.Uentry tag when is_genarg tag (ArgumentType wit_tactic) ->
     prtac LevelSome arg
-  | Extend.Uentryl (_, l) -> prtac LevelSome arg
+  | Extend.Uentryl (_, l) -> prtac (LevelLe l) arg
   | _ ->
     match arg with
     | TacGeneric (isquot,arg) ->
@@ -641,9 +641,8 @@ let pr_goal_selector ~toplevel s =
   let lseq = 4
   let ltactical = 3
   let lorelse = 2
-  let llet = 5
+  let llet = 5 (* in principle necessary only on the left-hand side of binary operators *)
   let lfun = 5
-  let lcomplete = 1
   let labstract = 3
   let lmatch = 1
   let latom = 0
@@ -1027,8 +1026,6 @@ let pr_goal_selector ~toplevel s =
               keyword "first" ++ spc () ++ pr_seq_body (pr_tac ltop) tl, llet
             | TacSolve tl ->
               keyword "solve" ++ spc () ++ pr_seq_body (pr_tac ltop) tl, llet
-            | TacComplete t ->
-              pr_tac (LevelLe lcomplete) t, lcomplete
             | TacSelect (s, tac) -> pr_goal_selector ~toplevel:false s ++ spc () ++ pr_tac ltop tac, ltactical
             | TacId l ->
               keyword "idtac" ++ prlist (pr_arg (pr_message_token pr.pr_name)) l, latom
@@ -1060,7 +1057,7 @@ let pr_goal_selector ~toplevel s =
               pr_with_comments ?loc (pr.pr_alias (level_of inherited) kn l), latom
           )
           in
-          if prec_less prec inherited then strm
+          if Notation.prec_less prec inherited then strm
           else str"(" ++ strm ++ str")"
 
         and pr_tacarg = function
@@ -1345,7 +1342,7 @@ let () =
     (make_constr_printer Printer.pr_econstr_n_env)
   ;
   Genprint.register_print0
-    wit_red_expr
+    Redexpr.wit_red_expr
     (lift_env (fun env sigma -> pr_red_expr env sigma (pr_constr_expr, pr_lconstr_expr, pr_or_by_notation pr_qualid, pr_constr_pattern_expr)))
     (lift_env (fun env sigma -> pr_red_expr env sigma
                   ((fun env sigma -> pr_and_constr_expr @@ pr_glob_constr_pptac env sigma),
