@@ -112,7 +112,7 @@ let option_assert_get o msg =
 (** Constructors for rawconstr *)
 open Glob_term
 
-let mkRHole = DAst.make @@ GHole (Evar_kinds.InternalHole, Namegen.IntroAnonymous)
+let mkRHole = DAst.make @@ GHole (GInternalHole)
 
 let rec mkRHoles n = if n > 0 then mkRHole :: mkRHoles (n - 1) else []
 let rec isRHoles cl = match cl with
@@ -122,7 +122,7 @@ let mkRApp f args = if args = [] then f else DAst.make @@ GApp (f, args)
 let mkRVar id = DAst.make @@ GRef (GlobRef.VarRef id,None)
 let mkRltacVar id = DAst.make @@ GVar (id)
 let mkRCast rc rt =  DAst.make @@ GCast (rc, Some DEFAULTcast, rt)
-let mkRType =  DAst.make @@ GSort (UAnonymous {rigid=true})
+let mkRType =  DAst.make @@ GSort (UAnonymous {rigid=UnivRigid})
 let mkRProp =  DAst.make @@ GSort (UNamed (None, [GProp, 0]))
 let mkRArrow rt1 rt2 = DAst.make @@ GProd (Anonymous, Explicit, rt1, rt2)
 let mkRConstruct c = DAst.make @@ GRef (GlobRef.ConstructRef c,None)
@@ -152,10 +152,10 @@ let intern_term ist env (_, c) = glob_constr ist env c
 
 let splay_open_constr env (sigma, c) =
   let t = Retyping.get_type_of env sigma c in
-  Reductionops.hnf_decompose_prod env sigma t
+  Reductionops.whd_decompose_prod env sigma t
 
 let isAppInd env sigma c =
-  let c = Reductionops.clos_whd_flags CClosure.all env sigma c in
+  let c = Reductionops.clos_whd_flags RedFlags.all env sigma c in
   let c, _ = EConstr.decompose_app sigma c in
   EConstr.isInd sigma c
 
@@ -353,7 +353,7 @@ let is_id_constr sigma c = match EConstr.kind sigma c with
 
 let red_product_skip_id env sigma c = match EConstr.kind sigma c with
   | App(hd,args) when Array.length args = 1 && is_id_constr sigma hd -> args.(0)
-  | _ -> try Tacred.red_product env sigma c with _ -> c
+  | _ -> match Tacred.red_product env sigma c with Some c -> c | None -> c
 
 let ssrevaltac ist gtac = Tacinterp.tactic_of_value ist gtac
 
@@ -460,13 +460,10 @@ let call_on_evar env sigma tac e =
     Exninfo.iraise (e, info)
 
 open Pp
-let pp _ = () (* FIXME *)
 module Intset = Evar.Set
 
 let abs_evars_pirrel env sigma0 (sigma, c0) =
-  pp(lazy(str"==PF_ABS_EVARS_PIRREL=="));
-  pp(lazy(str"c0= " ++ Printer.pr_econstr_env env sigma c0));
-  let c0 = Evarutil.nf_evar sigma0 (Evarutil.nf_evar sigma c0) in
+  let c0 = Evarutil.nf_evar sigma c0 in
   let nenv = env_size env in
   let abs_evar n k =
     let open EConstr in
@@ -477,22 +474,20 @@ let abs_evars_pirrel env sigma0 (sigma, c0) =
     | NamedDecl.LocalDef (x,b,t) -> mkNamedLetIn sigma x b t (mkArrow t x.binder_relevance c)
     | NamedDecl.LocalAssum (x,t) -> mkNamedProd sigma x t c in
     let t = Context.Named.fold_inside abs_dc ~init:concl dc in
-    Evarutil.nf_evar sigma0 (Evarutil.nf_evar sigma t) in
+    Evarutil.nf_evar sigma t
+  in
   let rec put evlist c = match EConstr.kind sigma c with
   | Evar (k, a) ->
     if List.mem_assoc k evlist || Evd.mem sigma0 k then evlist else
     let n = max 0 (SList.length a - nenv) in
-    let k_ty =
-      Retyping.get_sort_family_of
-        env sigma (Evd.evar_concl (Evd.find_undefined sigma k)) in
+    (* FIXME? this is not the right environment in general *)
+    let k_ty = Retyping.get_sort_family_of env sigma (Evd.evar_concl (Evd.find_undefined sigma k)) in
     let is_prop = k_ty = InProp in
-    let t = abs_evar n k in (k, (n, t, is_prop)) :: put evlist t
+    let t = abs_evar n k in
+    (k, (n, t, is_prop)) :: put evlist t
   | _ -> EConstr.fold sigma put evlist c in
   let evlist = put [] c0 in
   if evlist = [] then 0, c0 else
-  let pr_constr t = Printer.pr_econstr_env env sigma (Reductionops.nf_beta env sigma0 t) in
-  pp(lazy(str"evlist=" ++ pr_list (fun () -> str";")
-    (fun (k,_) -> Evar.print k) evlist));
   let evplist =
     let depev = List.fold_left (fun evs (_,(_,t,_)) ->
         Intset.union evs (Evarutil.undefined_evars_of_term sigma t)) Intset.empty evlist in
@@ -503,13 +498,11 @@ let abs_evars_pirrel env sigma0 (sigma, c0) =
       try
         let sigma = call_on_evar env sigma !ssrautoprop_tac i in
         List.filter (fun (j,_) -> j <> i) ev, evp, sigma
-      with _ -> ev, p::evp, sigma) (evlist, [], sigma) (List.rev evplist) in
+      with e when CErrors.noncritical e -> ev, p::evp, sigma) (evlist, [], sigma) (List.rev evplist) in
   let c0 = Evarutil.nf_evar sigma c0 in
-  let evlist =
-    List.map (fun (x,(y,t,z)) -> x,(y,Evarutil.nf_evar sigma t,z)) evlist in
-  let evplist =
-    List.map (fun (x,(y,t,z)) -> x,(y,Evarutil.nf_evar sigma t,z)) evplist in
-  pp(lazy(str"c0= " ++ pr_constr c0));
+  let nf (k, (n, t, p)) = (k, (n, Evarutil.nf_evar sigma t, p)) in
+  let evlist = List.map nf evlist in
+  let evplist = List.map nf evplist in
   let rec lookup k i = function
     | [] -> 0, 0
     | (k', (n,_,_)) :: evl -> if k = k' then i,n else lookup k (i + 1) evl in
@@ -526,26 +519,25 @@ let abs_evars_pirrel env sigma0 (sigma, c0) =
       let j = destRel sigma hd in
       mkApp (mkRel j, Array.of_list (List.map (Vars.lift (i-1)) extra_args @ args))
   | _ -> EConstr.map_with_binders sigma ((+) 1) (app extra_args) i c in
-  let rec loopP evlist c i = function
+  let rec loopP evlist accu i = function
+  | [] -> List.rev accu
   | (_, (n, t, _)) :: evl ->
     let t = get evlist (i - 1) t in
     let n = Name (Id.of_string (ssr_anon_hyp ^ string_of_int n)) in
-    loopP evlist (mkProd (make_annot n Sorts.Relevant, t, c)) (i - 1) evl
-  | [] -> c in
+    loopP evlist (RelDecl.LocalAssum (make_annot n Sorts.Relevant, t) :: accu) (i - 1) evl
+  in
   let rec loop c i = function
   | (_, (n, t, _)) :: evl ->
     let evs = Evarutil.undefined_evars_of_term sigma t in
     let t_evplist = List.filter (fun (k,_) -> Intset.mem k evs) evplist in
-    let t = loopP t_evplist (get t_evplist 1 t) 1 t_evplist in
+    let ctx_t = loopP t_evplist [] 1 t_evplist in
+    let t = EConstr.it_mkProd_or_LetIn (get t_evplist 1 t) ctx_t in
     let t = get evlist (i - 1) t in
-    let extra_args =
-      List.map (fun (k,_) -> mkRel (fst (lookup k i evlist)))
-        (List.rev t_evplist) in
+    let extra_args = List.rev_map (fun (k,_) -> mkRel (fst (lookup k i evlist))) t_evplist in
     let c = if extra_args = [] then c else app extra_args 1 c in
     loop (mkLambda (make_annot (mk_evar_name n) Sorts.Relevant, t, c)) (i - 1) evl
   | [] -> c in
   let res = loop (get evlist 1 c0) 1 evlist in
-  pp(lazy(str"res= " ++ pr_constr res));
   List.length evlist, res
 
 (* Strip all non-essential dependencies from an abstracted term, generating *)
@@ -556,7 +548,7 @@ let nb_evar_deps = function
     let s = Id.to_string id in
     if not (is_tagged evar_tag s) then 0 else
     let m = String.length evar_tag in
-    (try int_of_string (String.sub s m (String.length s - 1 - m)) with _ -> 0)
+    (try int_of_string (String.sub s m (String.length s - 1 - m)) with e when CErrors.noncritical e -> 0)
   | _ -> 0
 
 let type_id env sigma t = Id.of_string (Namegen.hdchar env sigma t)
@@ -719,11 +711,11 @@ open Util
 
 (** Constructors for constr_expr *)
 let mkCProp loc = CAst.make ?loc @@ CSort (UNamed (None, [CProp, 0]))
-let mkCType loc = CAst.make ?loc @@ CSort (UAnonymous {rigid=true})
+let mkCType loc = CAst.make ?loc @@ CSort (UAnonymous {rigid=UnivRigid})
 let mkCVar ?loc id = CAst.make ?loc @@ CRef (qualid_of_ident ?loc id, None)
 let rec mkCHoles ?loc n =
-  if n <= 0 then [] else (CAst.make ?loc @@ CHole (None, Namegen.IntroAnonymous)) :: mkCHoles ?loc (n - 1)
-let mkCHole loc = CAst.make ?loc @@ CHole (None, Namegen.IntroAnonymous)
+  if n <= 0 then [] else (CAst.make ?loc @@ CHole (None)) :: mkCHoles ?loc (n - 1)
+let mkCHole loc = CAst.make ?loc @@ CHole (None)
 let mkCLambda ?loc name ty t =  CAst.make ?loc @@
    CLambdaN ([CLocalAssum([CAst.make ?loc name], Default Explicit, ty)], t)
 let mkCArrow ?loc ty t = CAst.make ?loc @@
@@ -811,39 +803,55 @@ let dependent_apply_error =
  * Refine.refine function that handles type classes and evars but fails to
  * handle "dependently typed higher order evars".
  *
+ * Assumes that the type of [t] is a telescope with [n] leading product types.
+ * This is always the case as the only caller of applyn uses it with the output
+ * of abs_evars_pirrel which starts with n lambdas generated by abstraction.
+ * It also assumes [t] starts with [n] Lambda nodes.
+ *
  * Refiner.refiner that does not handle metas with a non ground type but works
  * with dependently typed higher order metas. *)
-let applyn ~with_evars ?beta ?(with_shelve=false) ?(first_goes_last=false) n t =
+let applyn ?(beta=false) ~with_evars ?(first_goes_last=false) n t =
+  Proofview.Goal.enter begin fun gl ->
+  let sigma = Proofview.Goal.sigma gl in
+  let env = Proofview.Goal.env gl in
+  let concl = Proofview.Goal.concl gl in
   if with_evars then
     let refine =
-      Proofview.Goal.enter begin fun gl ->
       Refine.refine ~typecheck:false begin fun sigma ->
-      let env = Proofview.Goal.env gl in
-      let concl = Proofview.Goal.concl gl in
-      let t, ty, args, sigma = saturate ?beta ~bi_types:true env sigma t n in
+      let rec saturate c args sigma n =
+        if n = 0 then args, sigma
+        else match EConstr.kind sigma c with
+        | Lambda (_, argty, c) ->
+          let argty = Reductionops.nf_betaiota env sigma (EConstr.Vars.substl args argty) in
+          let (sigma, x) = Evarutil.new_evar env sigma argty in
+          saturate c (x :: args) sigma (n-1)
+        | _ -> assert false
+      in
+      let _, ty = EConstr.decompose_prod_n_decls sigma n (Retyping.get_type_of env sigma t) in
+      let args, sigma = saturate t [] sigma n in
+      let ty = EConstr.Vars.substl args ty in
+      let args = Array.rev_of_list args in
+      let t = EConstr.mkApp (t, args) in
+      let t = if beta then Reductionops.whd_beta env sigma t else t in
       let sigma = unify_HO env sigma ty concl in
       (* Set our own set of goals. In theory saturate generates them in the
          right order, so we could just return sigma directly, but explicit is
          better than implicit. *)
       let sigma = Evd.push_future_goals (snd @@ Evd.pop_future_goals sigma) in
-      let fold sigma (_, e, _) = match EConstr.kind sigma e with
+      let fold sigma e = match EConstr.kind sigma e with
       | Evar (evk, _) -> Evd.declare_future_goal evk sigma
       | _ -> sigma
       in
-      let sigma = List.fold_left fold sigma args in
+      let sigma = Array.fold_left fold sigma args in
       (sigma, t)
       end
-      end
     in
     Tacticals.tclTHENLIST [
       refine;
-      Proofview.(if with_shelve then shelve_unifiable else tclUNIT ());
+      Proofview.shelve_unifiable;
       Proofview.(if first_goes_last then cycle 1 else tclUNIT ())
     ]
   else
-    Proofview.Goal.enter begin fun gl ->
-    let sigma = Proofview.Goal.sigma gl in
-    let env = Proofview.Goal.env gl in
     let sigma = Evd.push_future_goals sigma in
     let hyps = Environ.named_context_val env in
     let inst = EConstr.identity_subst_val hyps in
@@ -855,15 +863,13 @@ let applyn ~with_evars ?beta ?(with_shelve=false) ?(first_goes_last=false) n t =
           | Lambda (_, ty, bo) ->
             let () = if not (EConstr.Vars.closed0 sigma ty) then raise dependent_apply_error in
             let ty = Reductionops.nf_betaiota env sigma ty in
-            let src = Loc.tag Evar_kinds.GoalEvar in
-            let (sigma, evk) = Evarutil.new_pure_evar ~src ~typeclass_candidate:false hyps sigma ty in
+            let (sigma, evk) = Evarutil.new_pure_evar ~typeclass_candidate:false hyps sigma ty in
             loop sigma bo (evk :: args) (n - 1)
           | _ -> assert false
       in
       loop sigma t [] n
     in
     let _, sigma = Evd.pop_future_goals sigma in
-    pp(lazy(str"Refiner.refiner " ++ Printer.pr_econstr_env env sigma t));
 
     let map evk = Proofview.goal_with_state evk (Proofview.Goal.state gl) in
     let sgl = List.rev_map map args in
@@ -889,7 +895,7 @@ let refine_with ?(first_goes_last=false) ?beta ?(with_evars=true) oc =
   let uct = Evd.evar_universe_context (fst oc) in
   let n, oc = abs_evars_pirrel env sigma oc in
   Proofview.Unsafe.tclEVARS (Evd.set_universe_context sigma uct) <*>
-  Proofview.tclORELSE (applyn ~with_evars ~first_goes_last ~with_shelve:true ?beta n oc)
+  Proofview.tclORELSE (applyn ~with_evars ~first_goes_last ?beta n oc)
     (fun _ -> Proofview.tclZERO dependent_apply_error)
   end
 
@@ -1117,7 +1123,7 @@ let clr_of_wgen gen clrs = match gen with
 let reduct_in_concl ~check t = Tactics.reduct_in_concl ~cast:false ~check (t, DEFAULTcast)
 let unfold cl =
   Proofview.tclEVARMAP >>= fun sigma ->
-  let module R = Reductionops in let module F = CClosure.RedFlags in
+  let module R = Reductionops in let module F = RedFlags in
   let flags = F.mkflags [F.fBETA; F.fMATCH; F.fFIX; F.fCOFIX; F.fDELTA] in
   let fold accu c = F.red_add accu (F.fCONST (fst (EConstr.destConst sigma c))) in
   let flags = List.fold_left fold flags cl in
@@ -1194,7 +1200,7 @@ let tclFULL_BETAIOTA = Goal.enter begin fun gl ->
   let r, _ = Redexpr.reduction_of_red_expr (Goal.env gl)
     Genredexpr.(Lazy {
       rBeta=true; rMatch=true; rFix=true; rCofix=true;
-      rZeta=false; rDelta=false; rConst=[]}) in
+      rZeta=false; rDelta=false; rConst=[]; rStrength=Norm}) in
   Tactics.e_reduct_in_concl ~cast:false ~check:false (r,Constr.DEFAULTcast)
 end
 
@@ -1331,8 +1337,8 @@ let tacDEST_CONST c =
  * to change that behaviour in the standard unfold code *)
 let unprotecttac =
   tacMK_SSR_CONST "protect_term" >>= tacDEST_CONST >>= fun prot ->
-  let open CClosure.RedFlags in
-  let flags = red_add_transparent CClosure.allnolet TransparentState.empty in
+  let open RedFlags in
+  let flags = red_add_transparent allnolet TransparentState.empty in
   let flags = red_add flags (fCONST prot) in
   Tacticals.onClause (fun idopt ->
     let hyploc = Option.map (fun id -> id, InHyp) idopt in
