@@ -43,6 +43,7 @@ exception ReductionTacticError of reduction_tactic_error
 
 val is_evaluable : Environ.env -> evaluable_global_reference -> bool
 
+exception NotEvaluableRef of GlobRef.t
 val error_not_evaluable : GlobRef.t -> 'a
 
 val evaluable_of_global_reference :
@@ -51,13 +52,8 @@ val evaluable_of_global_reference :
 val global_of_evaluable_reference :
   evaluable_global_reference -> GlobRef.t
 
-exception Redelimination
-
-(** Red (raise user error if nothing reducible) *)
-val red_product : reduction_function
-
-(** Red (raise Redelimination if nothing reducible) *)
-val try_red_product : reduction_function
+(** Red (returns None if nothing reducible) *)
+val red_product : env -> evar_map -> constr -> constr option
 
 (** Simpl *)
 val simpl : reduction_function
@@ -69,6 +65,9 @@ val whd_simpl : reduction_function
    not immediately hide a non reducible fix or cofix *)
 val hnf_constr : reduction_function
 
+(** Variant of the above that does not perform nf-βι *)
+val hnf_constr0 : reduction_function
+
 (** Unfold *)
 val unfoldn :
   (occurrences * evaluable_global_reference) list ->  reduction_function
@@ -82,10 +81,11 @@ val pattern_occs : (occurrences * constr) list -> e_reduction_function
 (** Rem: Lazy strategies are defined in Reduction *)
 
 (** Call by value strategy (uses Closures) *)
-val cbv_norm_flags : CClosure.RedFlags.reds ->  reduction_function
+val cbv_norm_flags : RedFlags.reds -> strong:bool -> reduction_function
   val cbv_beta : reduction_function
   val cbv_betaiota : reduction_function
   val cbv_betadeltaiota :  reduction_function
+  val whd_compute :  reduction_function
   val compute :  reduction_function  (** = [cbv_betadeltaiota] *)
 
 (** [reduce_to_atomic_ind env sigma t] puts [t] in the form [t'=(I args)]
@@ -98,13 +98,19 @@ val reduce_to_atomic_ind : env ->  evar_map -> types -> (inductive * EInstance.t
    returns [I] and [t'] or fails with a user error *)
 val reduce_to_quantified_ind : env ->  evar_map -> types -> (inductive * EInstance.t) * types
 
+(** Same as {!reduce_to_quantified_ind} but more efficient because it does not
+    return the normalized type. *)
+val eval_to_quantified_ind : env -> evar_map -> types -> (inductive * EInstance.t)
+
 (** [reduce_to_quantified_ref env sigma ref t] try to put [t] in the form
-   [t'=(x1:A1)..(xn:An)(ref args)] and fails with user error if not possible *)
+    [t'=(x1:A1)..(xn:An)(ref args)]. When this is not possible, if [allow_failure]
+    is specified, [t] is unfolded until the point where this impossibility is plainly
+    visible. Otherwise, it fails with user error. *)
 val reduce_to_quantified_ref :
-  env ->  evar_map -> GlobRef.t -> types -> types
+  ?allow_failure:bool -> env ->  evar_map -> GlobRef.t -> types -> types
 
 val reduce_to_atomic_ref :
-  env ->  evar_map -> GlobRef.t -> types -> types
+  ?allow_failure:bool -> env ->  evar_map -> GlobRef.t -> types -> types
 
 val find_hnf_rectype :
   env ->  evar_map -> types -> (inductive * EInstance.t) * constr list
