@@ -55,13 +55,13 @@ let strip_params env sigma c =
   match EConstr.kind sigma c with
   | App (f, args) ->
     (match EConstr.kind sigma f with
-    | Const (cst,_) ->
-      (match Structures.PrimitiveProjections.find_opt cst with
-       | Some p ->
+    | Const cst ->
+      (match Structures.PrimitiveProjections.find_opt_with_relevance cst with
+       | Some (p,r) ->
          let p = Projection.make p false in
          let npars = Projection.npars p in
          if Array.length args > npars then
-           mkApp (mkProj (p, args.(npars)),
+           mkApp (mkProj (p, r, args.(npars)),
                   Array.sub args (npars+1) (Array.length args - (npars + 1)))
          else c
        | None -> c)
@@ -76,9 +76,9 @@ let clenv_refresh env sigma ctx clenv =
   let evd = Evd.meta_merge (Evd.meta_list clenv.evd) (Evd.clear_metas sigma) in
   match ctx with
   | Some ctx ->
-    let (subst, ctx) = UnivGen.fresh_universe_context_set_instance ctx in
+    let (subst, ctx) = UnivGen.fresh_sort_context_instance ctx in
     let emap c = Vars.subst_univs_level_constr subst c in
-    let evd = Evd.merge_context_set Evd.univ_flexible evd ctx in
+    let evd = Evd.merge_sort_context_set Evd.univ_flexible evd ctx in
     (* Only metas are mentioning the old universes. *)
     mk_clausenv env (Evd.map_metas emap evd) clenv.metas
       (emap clenv.templval)
@@ -297,9 +297,6 @@ let clenv_missing ce = clenv_dependent_gen true ce.env ce.evd (clenv_type ce)
 let clenv_unify ?(flags=default_unify_flags ()) cv_pb t1 t2 clenv =
   update_clenv_evd clenv (w_unify ~flags clenv.env clenv.evd cv_pb t1 t2)
 
-let clenv_unify_meta_type ?flags pb t mv clenv =
-  clenv_unify ?flags pb t (clenv_meta_type clenv.env clenv.evd mv) clenv
-
 let clenv_unify_meta_types ?(flags=default_unify_flags ()) clenv =
   update_clenv_evd clenv (w_unify_meta_types ~flags:flags clenv.env clenv.evd)
 
@@ -601,7 +598,7 @@ type proof =
 | RfHole of metavariable
 | RfGround of EConstr.t
 | RfApp of proof * proof list
-| RfProj of Projection.t * proof
+| RfProj of Projection.t * Sorts.relevance * proof
 
 exception NonLinear
 
@@ -622,9 +619,9 @@ let make_proof env sigma c =
     let args = Array.map_to_list (fun c -> make c) args in
     if is_ground f && List.for_all is_ground args then RfGround c
     else RfApp (f, args)
-  | Proj (p, a) ->
+  | Proj (p, r, a) ->
     let a = make a in
-    if is_ground a then RfGround c else RfProj (p, a)
+    if is_ground a then RfGround c else RfProj (p, r, a)
   | _ ->
     if occur_meta sigma c then error_unsupported_deep_meta ()
     else RfGround c
@@ -636,7 +633,7 @@ let rec as_constr = function
 | RfHole mv -> EConstr.mkMeta mv
 | RfGround c -> c
 | RfApp (f, args) -> EConstr.mkApp (as_constr f, Array.map_of_list as_constr args)
-| RfProj (p, c) -> EConstr.mkProj (p, as_constr c)
+| RfProj (p, r, c) -> EConstr.mkProj (p, r, as_constr c)
 
 (* Old style mk_goal primitive *)
 let mk_goal evars hyps concl =
@@ -681,9 +678,9 @@ let rec mk_refgoals env sigma goalacc conclty trm = match trm with
   let ((acc'',conclty',sigma), args) = mk_arggoals env sigma acc' hdty l in
   let ans = EConstr.applist (applicand, args) in
   (acc'', conclty', sigma, ans)
-| RfProj (p, c) ->
+| RfProj (p, r, c) ->
   let (acc',cty,sigma,c') = mk_refgoals env sigma goalacc None c in
-  let c = EConstr.mkProj (p, c') in
+  let c = EConstr.mkProj (p, r, c') in
   let ty = get_type_of env sigma c in
   (acc',ty,sigma,c)
 
@@ -831,19 +828,23 @@ let build_case_analysis env sigma (ind, u) params pred indices indarg dep knd =
 
   match projs with
   | None ->
-    let ci = make_case_info env ind relevance RegularStyle in
+    let ci = make_case_info env ind RegularStyle in
     let pbody =
       mkApp
         (pred,
           if dep then Context.Rel.instance mkRel 0 deparsign
           else Context.Rel.instance mkRel 1 (List.tl deparsign)) in
     let iv =
-      if Typeops.should_invert_case env ci then CaseInvert { indices = indices }
+      if Typeops.should_invert_case env relevance ci then CaseInvert { indices = indices }
       else NoInvert
     in
-    RealCase (ci, u, params, (pnas, pbody), iv, indarg)
+    RealCase (ci, u, params, ((pnas, pbody), relevance), iv, indarg)
   | Some ps ->
-    let args = Array.map (fun p -> mkProj (Projection.make p true, indarg)) ps in
+    let args = Array.map (fun (p,r) ->
+        let r = UVars.subst_instance_relevance (Unsafe.to_instance u) r in
+        mkProj (Projection.make p true, r, indarg))
+        ps
+    in
     PrimitiveEta args
 
 let case_pf ?(with_evars=false) ~dep (indarg, typ) =
@@ -910,11 +911,11 @@ let case_pf ?(with_evars=false) ~dep (indarg, typ) =
   let rec nf_betaiota c = EConstr.map sigma nf_betaiota (whd_betaiota env sigma c) in
   (* Call the legacy refiner on the result *)
   let arg = match body with
-  | RealCase (ci, u, pms, p, iv, c) ->
+  | RealCase (ci, u, pms, (p,r), iv, c) ->
     let c = nf_betaiota c in
     let pms = Array.map nf_betaiota pms in
     let p = on_snd nf_betaiota p in
-    Internal.Case ((ci, u, pms, p, iv, c), branches)
+    Internal.Case ((ci, u, pms, (p,r), iv, c), branches)
   | PrimitiveEta args ->
     let mv = new_meta () in
     let (ctx, t) = branches.(0) in
@@ -956,13 +957,13 @@ let fail_quick_unif_flags = {
 }
 
 (* let unifyTerms m n = walking (fun wc -> fst (w_Unify CONV m n [] wc)) *)
-let unify ?(flags=fail_quick_unif_flags) m =
+let unify ?(flags=fail_quick_unif_flags) ~cv_pb m =
   Proofview.Goal.enter begin fun gl ->
     let env = Tacmach.pf_env gl in
     let n = Tacmach.pf_concl gl in
     let evd = clear_metas (Tacmach.project gl) in
     try
-      let evd' = w_unify env evd CONV ~flags m n in
+      let evd' = w_unify env evd cv_pb ~flags m n in
         Proofview.Unsafe.tclEVARSADVANCE evd'
     with e when CErrors.noncritical e ->
       let info = Exninfo.reify () in
