@@ -22,7 +22,7 @@ let red_constant_body red_opt env sigma body = match red_opt with
     red env sigma body
 
 let warn_implicits_in_term =
-  CWarnings.create ~name:"implicits-in-term" ~category:"implicits"
+  CWarnings.create ~name:"implicits-in-term" ~category:CWarnings.CoreCategories.implicits
          (fun () ->
           strbrk "Implicit arguments declaration relies on type." ++ spc () ++
             strbrk "Discarding incompatible declaration in term.")
@@ -49,10 +49,10 @@ let protect_pattern_in_binder bl c ctypopt =
   if List.exists (function Constrexpr.CLocalPattern _ -> true | _ -> false) bl
   then
     let t = match ctypopt with
-      | None -> CAst.make ?loc:c.CAst.loc (Constrexpr.CHole (None,Namegen.IntroAnonymous,None))
+      | None -> CAst.make ?loc:c.CAst.loc (Constrexpr.CHole (None))
       | Some t -> t in
     let loc = Loc.merge_opt c.CAst.loc t.CAst.loc in
-    let c = CAst.make ?loc @@ Constrexpr.CCast (c, Constr.DEFAULTcast, t) in
+    let c = CAst.make ?loc @@ Constrexpr.CCast (c, Some Constr.DEFAULTcast, t) in
     let loc = match List.hd bl with
       | Constrexpr.CLocalAssum (a::_,_,_) | Constrexpr.CLocalDef (a,_,_) -> a.CAst.loc
       | Constrexpr.CLocalPattern {CAst.loc} -> loc
@@ -113,9 +113,9 @@ let interp_definition ~program_mode env evd impl_env bl red_option c ctypopt =
 
 let definition_using env evd ~body ~types ~using =
   let terms = Option.List.cons types [body] in
-  Option.map (fun using -> Proof_using.definition_using env evd ~using ~terms) using
+  Option.map (fun using -> Proof_using.definition_using env evd ~fixnames:[] ~using ~terms) using
 
-let do_definition ?hook ~name ?scope ~poly ?typing_flags ~kind ?using udecl bl red_option c ctypopt =
+let do_definition ?hook ~name ?scope ?clearbody ~poly ?typing_flags ~kind ?using ?deprecation udecl bl red_option c ctypopt =
   let program_mode = false in
   let env = Global.env() in
   let env = Environ.update_typing_flags ?typing_flags env in
@@ -127,12 +127,12 @@ let do_definition ?hook ~name ?scope ~poly ?typing_flags ~kind ?using udecl bl r
   let using = definition_using env evd ~body ~types ~using in
   let kind = Decls.IsDefinition kind in
   let cinfo = Declare.CInfo.make ~name ~impargs ~typ:types ?using () in
-  let info = Declare.Info.make ?scope ~kind ?hook ~udecl ~poly ?typing_flags () in
+  let info = Declare.Info.make ?scope ?clearbody ~kind ?hook ~udecl ~poly ?typing_flags ?deprecation () in
   let _ : Names.GlobRef.t =
     Declare.declare_definition ~info ~cinfo ~opaque:false ~body evd
   in ()
 
-let do_definition_program ?hook ~pm ~name ~scope ~poly ?typing_flags ~kind ?using udecl bl red_option c ctypopt =
+let do_definition_program ?hook ~pm ~name ~scope ?clearbody ~poly ?typing_flags ~kind ?using ?deprecation udecl bl red_option c ctypopt =
   let () = if not poly then udecl |> Option.iter (fun udecl ->
       if not udecl.UState.univdecl_extensible_instance
       || not udecl.UState.univdecl_extensible_constraints
@@ -152,6 +152,6 @@ let do_definition_program ?hook ~pm ~name ~scope ~poly ?typing_flags ~kind ?usin
   let term, typ, uctx, obls = Declare.Obls.prepare_obligation ~name ~body ~types evd in
   let pm, _ =
     let cinfo = Declare.CInfo.make ~name ~typ ~impargs ?using () in
-    let info = Declare.Info.make ~udecl ~scope ~poly ~kind ?hook ?typing_flags () in
+    let info = Declare.Info.make ~udecl ~scope ?clearbody ~poly ~kind ?hook ?typing_flags ?deprecation () in
     Declare.Obls.add_definition ~pm ~cinfo ~info ~term ~uctx obls
   in pm
