@@ -42,7 +42,8 @@ let short = ref false
 
 let () =
   declare_bool_option
-    { optdepr  = false;
+    { optstage = Summary.Stage.Interp;
+      optdepr  = None;
       optkey   = ["Short";"Module";"Printing"];
       optread  = (fun () -> !short) ;
       optwrite = ((:=) short) }
@@ -109,16 +110,16 @@ let print_fields envpar sigma cstrtypes =
         Printer.pr_lconstr_env envpar sigma c) fields) ++ str" }"
 
 let print_one_inductive env sigma isrecord mib ((_,i) as ind) =
-  let u = Univ.make_abstract_instance (Declareops.inductive_polymorphic_context mib) in
+  let u = UVars.make_abstract_instance (Declareops.inductive_polymorphic_context mib) in
   let mip = mib.mind_packets.(i) in
   let paramdecls = Inductive.inductive_paramdecls (mib,u) in
   let env_params, params = Namegen.make_all_rel_context_name_different env (Evd.from_env env) (EConstr.of_rel_context paramdecls) in
   let params = EConstr.Unsafe.to_rel_context params in
   let nparamdecls = Context.Rel.length params in
   let args = Context.Rel.instance_list mkRel 0 params in
-  let arity = hnf_prod_applist_assum env nparamdecls (build_ind_type ((mib,mip),u)) args in
+  let arity = hnf_prod_applist_decls env nparamdecls (build_ind_type ((mib,mip),u)) args in
   let cstrtypes = Inductive.type_of_constructors (ind,u) (mib,mip) in
-  let cstrtypes = Array.map (fun c -> hnf_prod_applist_assum env nparamdecls c args) cstrtypes in
+  let cstrtypes = Array.map (fun c -> hnf_prod_applist_decls env nparamdecls c args) cstrtypes in
   if isrecord then assert (Array.length cstrtypes = 1);
   let inst =
     if Declareops.inductive_is_polymorphic mib then
@@ -214,7 +215,7 @@ let nametab_register_body mp dir (l,body) =
         mib.mind_packets
 
 (* TODO only import printing-relevant objects (or find a way to print without importing) *)
-let import_module = Declaremods.import_module Libobject.unfiltered
+let import_module = Declaremods.Interp.import_module Libobject.unfiltered
 let process_module_binding = Declaremods.process_module_binding
 
 let nametab_register_module_body mp struc =
@@ -226,10 +227,10 @@ let nametab_register_module_body mp struc =
     List.iter (nametab_register_body mp DirPath.empty) struc
 
 let get_typ_expr_alg mtb = match mtb.mod_type_alg with
-  | Some (NoFunctor me) -> me
+  | Some (MENoFunctor me) -> me
   | _ -> raise Not_found
 
-let nametab_register_modparam mbid mtb =
+let nametab_register_modparam used mbid mtb =
   let id = MBId.to_id mbid in
   match mtb.mod_type with
   | MoreFunctor _ -> id (* functorial param : nothing to register *)
@@ -242,7 +243,7 @@ let nametab_register_modparam mbid mtb =
     with e when CErrors.noncritical e ->
       (* Otherwise, we try to play with the nametab ourselves *)
       let mp = MPbound mbid in
-      let check id = Nametab.exists_module (DirPath.make [id]) in
+      let check id = Id.Set.mem id used || Nametab.exists_module (DirPath.make [id]) in
       let id = Namegen.next_ident_away_from id check in
       let dir = DirPath.make [id] in
       nametab_register_dir mp;
@@ -339,19 +340,20 @@ let print_mod_expr env mp locals = function
         (str"(" ++ prlist_with_sep spc (print_modpath locals) lapp ++ str")")
   | MEwith _ -> assert false (* No 'with' syntax for modules *)
 
-let rec print_functor fty fatom is_type extent env mp locals = function
+let rec print_functor fty fatom is_type extent env mp used locals = function
   | NoFunctor me -> fatom is_type extent env mp locals me
   | MoreFunctor (mbid,mtb1,me2) ->
-      let id = nametab_register_modparam mbid mtb1 in
+      let id = nametab_register_modparam !used mbid mtb1 in
+      let () = used := Id.Set.add id !used in
       let mp1 = MPbound mbid in
-      let pr_mtb1 = fty extent env mp1 locals mtb1 in
+      let pr_mtb1 = fty extent env mp1 used locals mtb1 in
       let env' = Modops.add_module_type mp1 mtb1 env in
       let locals' = (mbid, get_new_id locals (MBId.to_id mbid))::locals in
       let kwd = if is_type then "Funsig" else "Functor" in
       hov 2
         (keyword kwd ++ spc () ++
          str "(" ++ Id.print id ++ str ":" ++ pr_mtb1 ++ str ")" ++
-         spc() ++ print_functor fty fatom is_type extent env' mp locals' me2)
+         spc() ++ print_functor fty fatom is_type extent env' mp used locals' me2)
 
 let rec print_expression x =
   print_functor
@@ -361,20 +363,22 @@ let rec print_expression x =
 and print_signature x =
   print_functor print_modtype print_structure x
 
-and print_modtype extent env mp locals mtb = match mtb.mod_type_alg with
-  | Some me -> print_expression true extent env mp locals me
-  | None -> print_signature true extent env mp locals mtb.mod_type
+and print_modtype extent env mp used locals mtb = match mtb.mod_type_alg with
+  | Some me ->
+    let me = Modops.annotate_module_expression me mtb.mod_type in
+    print_expression true extent env mp used locals me
+  | None -> print_signature true extent env mp used locals mtb.mod_type
 
 (** Since we might play with nametab above, we should reset to prior
     state after the printing *)
 
 let print_expression' is_type extent env mp me =
   Vernacstate.System.protect
-    (fun e -> print_expression is_type extent env mp [] e) me
+    (fun e -> print_expression is_type extent env mp (ref Id.Set.empty) [] e) me
 
 let print_signature' is_type extent env mp me =
   Vernacstate.System.protect
-    (fun e -> print_signature is_type extent env mp [] e) me
+    (fun e -> print_signature is_type extent env mp (ref Id.Set.empty) [] e) me
 
 let unsafe_print_module extent env mp with_body mb =
   let name = print_modpath [] mp in
@@ -382,13 +386,19 @@ let unsafe_print_module extent env mp with_body mb =
   let body = match with_body, mb.mod_expr with
     | false, _
     | true, Abstract -> mt()
-    | _, Algebraic me -> pr_equals ++ print_expression' false extent env mp me
-    | _, Struct sign -> pr_equals ++ print_signature' false extent env mp sign
+    | _, Algebraic me ->
+      let me = Modops.annotate_module_expression me mb.mod_type in
+      pr_equals ++ print_expression' false extent env mp me
+    | _, Struct sign ->
+      let sign = Modops.annotate_struct_body sign mb.mod_type in
+      pr_equals ++ print_signature' false extent env mp sign
     | _, FullStruct -> pr_equals ++ print_signature' false extent env mp mb.mod_type
   in
   let modtype = match mb.mod_expr, mb.mod_type_alg with
     | FullStruct, _ -> mt ()
-    | _, Some ty -> brk (1,1) ++ str": " ++ print_expression' true extent env mp ty
+    | _, Some ty ->
+      let ty = Modops.annotate_module_expression ty mb.mod_type in
+      brk (1,1) ++ str": " ++ print_expression' true extent env mp ty
     | _, _ -> brk (1,1) ++ str": " ++ print_signature' true extent env mp mb.mod_type
   in
   hv 0 (keyword "Module" ++ spc () ++ name ++ modtype ++ body)
