@@ -28,17 +28,17 @@ module RelDecl = Context.Rel.Declaration
 
 (********** definition d'un record (structure) **************)
 
-let typeclasses_strict =
+let { Goptions.get = typeclasses_strict } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Typeclasses";"Strict";"Resolution"]
     ~value:false
+    ()
 
-let typeclasses_unique =
+let { Goptions.get = typeclasses_unique } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Typeclasses";"Unique";"Instances"]
     ~value:false
+    ()
 
 let interp_fields_evars env sigma ~ninds ~nparams impls_env nots l =
   let _, sigma, impls, newfs, _ =
@@ -85,7 +85,7 @@ let interp_fields_evars env sigma ~ninds ~nparams impls_env nots l =
 
 let check_anonymous_type ind =
   match ind with
-  | { CAst.v = CSort (Glob_term.UAnonymous {rigid=true}) } -> true
+  | { CAst.v = CSort (Glob_term.UAnonymous {rigid=UnivRigid}) } -> true
   | _ -> false
 
 let error_parameters_must_be_named bk {CAst.loc; v=name} =
@@ -100,7 +100,7 @@ let check_parameters_must_be_named = function
   | CLocalAssum (ls, bk, _ce) ->
     List.iter (error_parameters_must_be_named bk) ls
   | CLocalPattern {CAst.loc} ->
-    Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed in record parameters")
+    Loc.raise ?loc (Gramlib.Grammar.Error "pattern with quote not allowed in record parameters")
 
 (** [DataI.t] contains the information used in record interpretation,
    it is a strict subset of [Ast.t] thus this should be
@@ -110,7 +110,7 @@ module DataI = struct
     { name : Id.t
     ; arity : Constrexpr.constr_expr option
     (** declared sort for the record  *)
-    ; nots : Metasyntax.where_decl_notation list list
+    ; nots : Metasyntax.notation_interpretation_decl list list
     (** notations for fields *)
     ; fs : Vernacexpr.local_decl_expr list
     }
@@ -120,8 +120,7 @@ end
    type-inference *)
 module DataR = struct
   type t =
-    { min_univ : Sorts.t
-    ; arity : Constr.t
+    { arity : Constr.t
     ; implfs : Impargs.manual_implicits list
     ; fields : Constr.rel_declaration list
     }
@@ -147,30 +146,33 @@ module Data = struct
   }
 end
 
+(** Is [s] a single local level (type or qsort)? If so return it. *)
+let is_sort_variable sigma s =
+  match EConstr.ESorts.kind sigma s with
+  | SProp | Prop | Set -> None
+  | Type u | QSort (_, u) -> match Univ.Universe.level u with
+    | None -> None
+    | Some l ->
+      if Univ.Level.Set.mem l (fst (Evd.universe_context_set sigma))
+      then Some l
+      else None
+
 let build_type_telescope newps env0 sigma { DataI.arity; _ } = match arity with
   | None ->
-    let uvarkind = Evd.univ_flexible_alg in
-    let sigma, s = Evd.new_sort_variable uvarkind sigma in
+    let sigma, s = Evd.new_sort_variable Evd.univ_flexible_alg sigma in
+    sigma, (EConstr.mkSort s, s)
+  | Some { CAst.v = CSort (Glob_term.UAnonymous {rigid=UnivRigid}); loc } ->
+    (* special case: the user wrote ": Type". We want to allow it to become algebraic
+       (and Prop but that may change in the future) *)
+    let sigma, s = Evd.new_sort_variable ?loc UState.univ_flexible_alg sigma in
     sigma, (EConstr.mkSort s, s)
   | Some t ->
     let env = EConstr.push_rel_context newps env0 in
-    let poly =
-      match t with
-      | { CAst.v = CSort (Glob_term.UAnonymous {rigid=true}) } -> true | _ -> false in
     let impls = Constrintern.empty_internalization_env in
     let sigma, s = Constrintern.interp_type_evars ~program_mode:false env sigma ~impls t in
     let sred = Reductionops.whd_allnolet env sigma s in
     (match EConstr.kind sigma sred with
-     | Sort s' ->
-       let s' = EConstr.ESorts.kind sigma s' in
-       (if poly then
-          match Evd.is_sort_variable sigma s' with
-          | Some l ->
-            let sigma = Evd.make_flexible_variable sigma ~algebraic:true l in
-            sigma, (s, s')
-          | None ->
-            sigma, (s, s')
-        else sigma, (s, s'))
+     | Sort s' -> (sigma, (s, s'))
      | _ -> user_err ?loc:(constr_loc t) (str"Sort expected."))
 
 type tc_result =
@@ -181,6 +183,26 @@ type tc_result =
   * Constr.rel_context
   * DataR.t list
 
+let def_class_levels ~def env_ar sigma aritysorts ctors =
+  let s, ctor = match aritysorts, ctors with
+    | [s], [_,ctor] -> begin match ctor with
+        | [LocalAssum (_,t)] -> s, t
+        | _ -> assert false
+      end
+    | _ -> CErrors.user_err Pp.(str "Mutual definitional classes are not supported.")
+  in
+  let ctor_sort = Retyping.get_sort_of env_ar sigma ctor in
+  let is_prop_ctor = EConstr.ESorts.is_prop sigma ctor_sort in
+  let sigma = Evd.set_leq_sort env_ar sigma ctor_sort s in
+  if Option.cata (Evd.is_flexible_level sigma) false (is_sort_variable sigma s)
+  && is_prop_ctor
+  then (* We assume that the level in aritysort is not constrained
+          and clear it, if it is flexible *)
+    let sigma = Evd.set_eq_sort env_ar sigma EConstr.ESorts.set s in
+    (sigma, [EConstr.mkProp])
+  else
+    sigma, [EConstr.mkSort s]
+
 (* ps = parameter list *)
 let typecheck_params_and_fields def poly udecl ps (records : DataI.t list) : tc_result =
   let env0 = Global.env () in
@@ -189,15 +211,16 @@ let typecheck_params_and_fields def poly udecl ps (records : DataI.t list) : tc_
      any Set <= i constraint for universes that might actually be instantiated with Prop. *)
   let is_template =
     List.exists (fun { DataI.arity; _} -> Option.cata check_anonymous_type true arity) records in
-  let env0 = if not poly && is_template then Environ.set_universes_lbound env0 UGraph.Bound.Prop else env0 in
+  let env0 = if not poly && not def && is_template then Environ.set_universes_lbound env0 UGraph.Bound.Prop else env0 in
   let sigma, decl, variances = Constrintern.interp_cumul_univ_decl_opt env0 udecl in
   let () = List.iter check_parameters_must_be_named ps in
   let sigma, (impls_env, ((_env1,newps), imps)) =
     Constrintern.interp_context_evars ~program_mode:false env0 sigma ps in
   let sigma, typs =
     List.fold_left_map (build_type_telescope newps env0) sigma records in
-  let arities = List.map (fun (typ, _) -> EConstr.it_mkProd_or_LetIn typ newps) typs in
-  let relevances = List.map (fun (_,s) -> Sorts.relevance_of_sort s) typs in
+  let typs, aritysorts = List.split typs in
+  let arities = List.map (fun typ -> EConstr.it_mkProd_or_LetIn typ newps) typs in
+  let relevances = List.map (fun s -> EConstr.ESorts.relevance_of_sort sigma s) aritysorts in
   let fold accu { DataI.name; _ } arity r =
     EConstr.push_rel (LocalAssum (make_annot (Name name) r,arity)) accu in
   let env_ar = EConstr.push_rel_context newps (List.fold_left3 fold env0 records arities relevances) in
@@ -214,31 +237,37 @@ let typecheck_params_and_fields def poly udecl ps (records : DataI.t list) : tc_
   let (sigma, data) = List.fold_left_map fold sigma records in
   let sigma =
     Pretyping.solve_remaining_evars Pretyping.all_and_fail_flags env_ar sigma in
-  let fold sigma (typ, sort) (_, newfs) =
-    let univ = ComInductive.Internal.compute_constructor_level env_ar sigma newfs in
-    let univ = if Sorts.is_sprop sort then univ else if Sorts.is_sprop univ then Sorts.prop else univ in
-      if not def && is_impredicative_sort env0 sort then
-        sigma, (sort, typ)
-      else
-        let sigma = Evd.set_leq_sort env_ar sigma univ sort in
-        if Sorts.is_small univ &&
-           Option.cata (Evd.is_flexible_level sigma) false (Evd.is_sort_variable sigma sort) then
-           (* We can assume that the level in aritysort is not constrained
-               and clear it, if it is flexible *)
-   Evd.set_eq_sort env_ar sigma Sorts.set sort, (univ, EConstr.mkSort univ)
-        else sigma, (univ, typ)
+  let sigma, typs =
+    if def then def_class_levels ~def env_ar sigma aritysorts data
+    else (* each inductive has one constructor *)
+      let ctors = List.map (fun (_,newfs) -> [newfs]) data in
+      ComInductive.Internal.inductive_levels env_ar sigma typs ctors
   in
-  let (sigma, typs) = List.fold_left2_map fold sigma typs data in
   (* TODO: Have this use Declaredef.prepare_definition *)
-  let sigma, (newps, ans) = Evarutil.finalize sigma (fun nf ->
-      let newps = List.map (RelDecl.map_constr_het nf) newps in
-      let map (implfs, fields) (min_univ, typ) =
-        let fields = List.map (RelDecl.map_constr_het nf) fields in
-        let arity = nf typ in
-        { DataR.min_univ; arity; implfs; fields }
-      in
-      let ans = List.map2 map data typs in
-      newps, ans)
+  let sigma, (newps, ans) =
+    (* too complex for Evarutil.finalize as we normalize non-constr *)
+    let sigma = Evd.minimize_universes sigma in
+    let uvars = ref Univ.Level.Set.empty in
+    let nf c =
+      let _, varsc = EConstr.universes_of_constr sigma c in
+      let c = EConstr.to_constr sigma c in
+      uvars := Univ.Level.Set.union !uvars varsc;
+      c
+    in
+    let nf_rel r = Evarutil.nf_relevance sigma r in
+    let map_decl = function
+    | LocalAssum (na, t) -> LocalAssum (UnivSubst.nf_binder_annot nf_rel na, nf t)
+    | LocalDef (na, c, t) -> LocalDef (UnivSubst.nf_binder_annot nf_rel na, nf c, nf t)
+    in
+    let newps = List.map map_decl newps in
+    let map (implfs, fields) typ =
+      let fields = List.map map_decl fields in
+      let arity = nf typ in
+      { DataR.arity; implfs; fields }
+    in
+    let ans = List.map2 map data typs in
+    let sigma = Evd.restrict_universe_context sigma !uvars in
+    sigma, (newps, ans)
   in
   let univs = Evd.check_univ_decl ~poly sigma decl in
   let ce t = Pretyping.check_evars env0 sigma (EConstr.of_constr t) in
@@ -250,9 +279,20 @@ type record_error =
   | BadTypedProj of Id.t * env * Type_errors.type_error
 
 let warn_cannot_define_projection =
-  CWarnings.create ~name:"cannot-define-projection" ~category:"records"
+  CWarnings.create ~name:"cannot-define-projection" ~category:CWarnings.CoreCategories.records
          (fun msg -> hov 0 msg)
 
+type arity_error =
+  | NonInformativeToInformative
+  | StrongEliminationOnNonSmallType
+
+let error_elim_explain kp ki =
+  let open Sorts in
+  match kp,ki with
+  | (InType | InSet), InProp -> Some NonInformativeToInformative
+  | InType, InSet -> Some StrongEliminationOnNonSmallType (* if Set impredicative *)
+  | _ -> None
+
 (* If a projection is not definable, we throw an error if the user
 asked it to be a coercion or instance. Otherwise, we just print an info
 message. The user might still want to name the field of the record. *)
@@ -264,20 +304,28 @@ let warning_or_error ~info flags indsp err =
            strbrk" cannot be defined because the projection" ++ str s ++ spc () ++
            prlist_with_sep pr_comma Id.print projs ++ spc () ++ str have ++
            strbrk " not defined.")
-    | BadTypedProj (fi,_ctx,te) ->
-        match te with
-          | ElimArity (_, _, Some (_, _, (InType | InSet), InProp)) ->
+    | BadTypedProj (fi,env,te) ->
+      let err = match te with
+        | ElimArity (_, _, Some (_, s)) ->
+          error_elim_explain (Sorts.family s)
+            (Inductiveops.elim_sort (Global.lookup_inductive indsp))
+        | _ -> None
+      in
+        match err with
+          | Some NonInformativeToInformative ->
               (Id.print fi ++
                 strbrk" cannot be defined because it is informative and " ++
                 Printer.pr_inductive (Global.env()) indsp ++
                 strbrk " is not.")
-          | ElimArity (_, _, Some (_, _, InType, InSet)) ->
+          | Some StrongEliminationOnNonSmallType ->
               (Id.print fi ++
                 strbrk" cannot be defined because it is large and " ++
                 Printer.pr_inductive (Global.env()) indsp ++
                 strbrk " is not.")
-          | _ ->
-              (Id.print fi ++ strbrk " cannot be defined because it is not typable.")
+          | None ->
+            (Id.print fi ++ str " cannot be defined because it is not typable:" ++ spc() ++
+             Himsg.explain_type_error env (Evd.from_env env)
+               (Type_errors.map_ptype_error EConstr.of_constr te))
   in
   if flags.Data.pf_coercion || flags.Data.pf_instance then user_err ~info st;
   warn_cannot_define_projection (hov 0 st)
@@ -330,11 +378,11 @@ let instantiate_possibly_recursive_type ind u ntypes paramdecls fields =
    or a typeclass instance according to [flags]. *)
 (* remove the last argument (it will become alway true) after deprecation phase
    (started in 8.17, c.f. https://github.com/coq/coq/pull/16230) *)
-let declare_proj_coercion_instance ~flags ref from ~poly ~with_coercion =
+let declare_proj_coercion_instance ~flags ref from ~with_coercion =
   if with_coercion && flags.Data.pf_coercion then begin
     let cl = ComCoercion.class_of_global from in
     let local = flags.Data.pf_locality = Goptions.OptLocal in
-    ComCoercion.try_add_new_coercion_with_source ref ~local ~poly ~nonuniform:false ~reversible:flags.Data.pf_reversible ~source:cl
+    ComCoercion.try_add_new_coercion_with_source ref ~local ~reversible:flags.Data.pf_reversible ~source:cl
   end;
   if flags.Data.pf_instance then begin
     let env = Global.env () in
@@ -364,18 +412,17 @@ let build_named_proj ~primitive ~flags ~poly ~univs ~uinstance ~kind env paramde
       (* [ccl] is defined in context [params;x:rp] *)
       (* [ccl'] is defined in context [params;x:rp;x:rp] *)
       if primitive then
-        let proj_relevant = match rci with Sorts.Irrelevant -> false | Sorts.Relevant -> true in
         let p = Projection.Repr.make indsp
-            ~proj_relevant ~proj_npars:mib.mind_nparams ~proj_arg:i (Label.of_id fid) in
-        mkProj (Projection.make p true, mkRel 1), Some p
+            ~proj_npars:mib.mind_nparams ~proj_arg:i (Label.of_id fid) in
+        mkProj (Projection.make p true, rci, mkRel 1), Some (p,rci)
       else
         let ccl' = liftn 1 2 ccl in
         let p = mkLambda (x, lift 1 rp, ccl') in
         let branch = it_mkLambda_or_LetIn (mkRel nfi) lifted_fields in
-        let ci = Inductiveops.make_case_info env indsp rci LetStyle in
+        let ci = Inductiveops.make_case_info env indsp LetStyle in
         (* Record projections are always NoInvert because they're at
            constant relevance *)
-        mkCase (Inductive.contract_case env (ci, p, NoInvert, mkRel 1, [|branch|])), None
+        mkCase (Inductive.contract_case env (ci, (p, rci), NoInvert, mkRel 1, [|branch|])), None
   in
   let proj = it_mkLambda_or_LetIn (mkLambda (x,rp,body)) paramdecls in
   let projtyp = it_mkProd_or_LetIn (mkProd (x,rp,ccl)) paramdecls in
@@ -393,9 +440,9 @@ let build_named_proj ~primitive ~flags ~poly ~univs ~uinstance ~kind env paramde
   in
   Declare.definition_message fid;
   let term = match p_opt with
-    | Some p ->
+    | Some (p,r) ->
       let _ = DeclareInd.declare_primitive_projection p kn in
-      mkProj (Projection.make p false,mkRel 1)
+      mkProj (Projection.make p false, r, mkRel 1)
     | None ->
       let proj_args = (*Rel 1 refers to "x"*) paramargs@[mkRel 1] in
       match decl with
@@ -404,7 +451,7 @@ let build_named_proj ~primitive ~flags ~poly ~univs ~uinstance ~kind env paramde
   in
   let refi = GlobRef.ConstRef kn in
   Impargs.maybe_declare_manual_implicits false refi impls;
-  declare_proj_coercion_instance ~flags refi (GlobRef.IndRef indsp) ~poly ~with_coercion:true;
+  declare_proj_coercion_instance ~flags refi (GlobRef.IndRef indsp) ~with_coercion:true;
   let i = if is_local_assum decl then i+1 else i in
   (Some kn, i, Projection term::subst)
 
@@ -441,14 +488,14 @@ let declare_projections indsp univs ?(kind=Decls.StructureComponent) inhabitant_
   let (mib,mip) = Global.lookup_inductive indsp in
   let poly = Declareops.inductive_is_polymorphic mib in
   let uinstance = match fst univs with
-    | Polymorphic_entry uctx -> Univ.UContext.instance uctx
-    | Monomorphic_entry -> Univ.Instance.empty
+    | Polymorphic_entry uctx -> UVars.UContext.instance uctx
+    | Monomorphic_entry -> UVars.Instance.empty
   in
   let paramdecls = Inductive.inductive_paramdecls (mib, uinstance) in
   let r = mkIndU (indsp,uinstance) in
   let rp = applist (r, Context.Rel.instance_list mkRel 0 paramdecls) in
   let paramargs = Context.Rel.instance_list mkRel 1 paramdecls in (*def in [[params;x:rp]]*)
-  let x = make_annot (Name inhabitant_id) mip.mind_relevance in
+  let x = make_annot (Name inhabitant_id) (Inductive.relevance_of_ind_body mip uinstance) in
   let fields = instantiate_possibly_recursive_type (fst indsp) uinstance mib.mind_ntypes paramdecls fields in
   let lifted_fields = Vars.lift_rel_context 1 fields in
   let primitive =
@@ -541,6 +588,7 @@ module Record_decl = struct
     ubinders : UnivNames.universe_binders;
     projections_kind : Decls.definition_object_kind;
     poly : bool;
+    indlocs : Loc.t option list;
   }
 end
 
@@ -570,12 +618,15 @@ let check_unique_names records =
       Vernacexpr.AssumExpr({CAst.v=Name id},_,_) -> id::acc
     | Vernacexpr.DefExpr ({CAst.v=Name id},_,_,_) -> id::acc
     | _ -> acc in
+  let indlocs =
+    records |> List.map (fun { Ast.name; _ } -> name ) in
+  let fields_names =
+    records |> List.fold_left (fun acc { Ast.cfs; _ } ->
+      List.fold_left extract_name acc cfs) [] in
   let allnames =
-    List.fold_left (fun acc { Ast.name; cfs; _ } ->
-      name.CAst.v :: (List.fold_left extract_name acc cfs)) [] records
-  in
+    fields_names @ (indlocs |> List.map (fun x -> x.CAst.v)) in
   match List.duplicates Id.equal allnames with
-  | [] -> ()
+  | [] -> List.map (fun x -> x.CAst.loc) indlocs
   | id :: _ -> user_err (str "Two objects have the same name" ++ spc () ++ quote (Id.print id) ++ str ".")
 
 type kind_class = NotClass | RecordClass | DefClass
@@ -620,17 +671,19 @@ let implicits_of_context ctx =
 (* deprecated in 8.16, to be removed at the end of the deprecation phase
    (c.f., https://github.com/coq/coq/pull/15802 ) *)
 let warn_future_coercion_class_constructor =
-  CWarnings.create ~name:"future-coercion-class-constructor" ~category:"records"
+  CWarnings.create ~name:"future-coercion-class-constructor" ~category:Deprecation.Version.v8_16
     ~default:CWarnings.AsError
     Pp.(fun () -> str "'Class >' currently does nothing. Use 'Class' instead.")
 
 (* deprecated in 8.17, to be removed at the end of the deprecation phase
    (c.f., https://github.com/coq/coq/pull/16230 ) *)
 let warn_future_coercion_class_field =
-  CWarnings.create ~name:"future-coercion-class-field" ~category:"records" Pp.(fun definitional ->
+  CWarnings.create ~name:"future-coercion-class-field" ~category:Deprecation.Version.v8_17
+    Pp.(fun definitional ->
     strbrk "A coercion will be introduced instead of an instance in future versions when using ':>' in 'Class' declarations. "
-    ++ strbrk "Replace ':>' with '::' (or use '#[global] Existing Instance field.' for compatibility with Coq < 8.17). Beware that the default locality for '::' is #[export], as opposed to #[global] for ':>' currently."
-    ++ strbrk (if definitional then "" else " Add an explicit #[global] attribute to the field if you need to keep the current behavior. For example: \"Class foo := { #[global] field :: bar }.\""))
+    ++ strbrk "Replace ':>' with '::' (or use '#[global] Existing Instance field.' for compatibility with Coq < 8.18). Beware that the default locality for '::' is #[export], as opposed to #[global] for ':>' currently."
+    ++ strbrk (if definitional then " Add an explicit #[global] attribute if you need to keep the current behavior. For example: \"Class foo := #[global] baz :: bar.\""
+               else " Add an explicit #[global] attribute to the field if you need to keep the current behavior. For example: \"Class foo := { #[global] field :: bar }.\""))
 
 let check_proj_flags kind rf =
   let open Vernacexpr in
@@ -672,7 +725,7 @@ let check_proj_flags kind rf =
    (started in 8.17)
    (c.f., https://github.com/coq/coq/pull/16230 ) *)
 let pre_process_structure ?(definitional=false) udecl kind ~poly (records : Ast.t list) =
-  let () = check_unique_names records in
+  let indlocs = check_unique_names records in
   let () = check_priorities kind records in
   let ps, data = extract_record_data records in
   let impargs, univs, variances, params, data =
@@ -707,9 +760,9 @@ let pre_process_structure ?(definitional=false) udecl kind ~poly (records : Ast.
   let data = List.map2 map data records in
   let projections_kind =
     Decls.(match kind_class kind with NotClass -> StructureComponent | _ -> Method) in
-  impargs, params, univs, variances, projections_kind, data
+  impargs, params, univs, variances, projections_kind, data, indlocs
 
-let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind data =
+let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind ~indlocs data =
   let nparams = List.length params in
   let (univs, ubinders) = univs in
   let poly, projunivs =
@@ -731,8 +784,8 @@ let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj i
   let blocks = List.mapi mk_block data in
   let ind_univs, global_univ_decls = match blocks, data with
   | [entry], [data] ->
-    let concl = Some data.Data.rdata.DataR.min_univ in
     let env_ar_params = Environ.push_rel_context params (Global.env ()) in
+    let concl = Some (snd (Reduction.dest_arity env_ar_params data.rdata.arity)) in
     ComInductive.compute_template_inductive ~user_template:template
       ~env_ar_params ~ctx_params:params ~univ_entry:univs entry concl
   | _ ->
@@ -767,25 +820,29 @@ let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj i
   in
   let impls = List.map (fun _ -> impargs, []) data in
   let open Record_decl in
-  { mie; primitive_proj; impls; globnames; global_univ_decls; projunivs; ubinders; projections_kind; poly; records = data }
+  { mie; primitive_proj; impls; globnames; global_univ_decls; projunivs;
+    ubinders; projections_kind; poly; records = data;
+    indlocs;
+  }
 
 
 let interp_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records =
-  let impargs, params, univs, variances, projections_kind, data =
+  assert (kind <> Vernacexpr.Class true);
+  let impargs, params, univs, variances, projections_kind, data, indlocs =
     pre_process_structure udecl kind ~poly records in
-  interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind data
+  interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind ~indlocs data
 
-let declare_structure { Record_decl.mie; primitive_proj; impls; globnames; global_univ_decls; projunivs; ubinders; projections_kind; poly; records } =
-  Option.iter (DeclareUctx.declare_universe_context ~poly:false) global_univ_decls;
+let declare_structure { Record_decl.mie; primitive_proj; impls; globnames; global_univ_decls; projunivs; ubinders; projections_kind; poly; records; indlocs } =
+  Option.iter (Global.push_context_set ~strict:true) global_univ_decls;
   let kn = DeclareInd.declare_mutual_inductive_with_eliminations mie globnames impls
-      ~primitive_expected:primitive_proj
+      ~primitive_expected:primitive_proj ~indlocs
   in
   let map i { Data.is_coercion; proj_flags; rdata = { DataR.implfs; fields; _}; inhabitant_id; _ } =
     let rsp = (kn, i) in (* This is ind path of idstruc *)
     let cstr = (rsp, 1) in
     let projections = declare_projections rsp (projunivs,ubinders) ~kind:projections_kind inhabitant_id proj_flags implfs fields in
     let build = GlobRef.ConstructRef cstr in
-    let () = if is_coercion then ComCoercion.try_add_new_coercion build ~local:false ~poly ~nonuniform:false ~reversible:true in
+    let () = if is_coercion then ComCoercion.try_add_new_coercion build ~local:false ~reversible:true in
     let struc = Structure.make (Global.env ()) rsp projections in
     let () = declare_structure_entry struc in
     GlobRef.IndRef rsp
@@ -819,9 +876,9 @@ let declare_class_constant ~univs paramimpls params data =
   in
   let inst, univs = match univs with
     | UState.Monomorphic_entry _, ubinders ->
-      Univ.Instance.empty, (UState.Monomorphic_entry Univ.ContextSet.empty, ubinders)
+      UVars.Instance.empty, (UState.Monomorphic_entry Univ.ContextSet.empty, ubinders)
     | UState.Polymorphic_entry uctx, _ ->
-      Univ.UContext.instance uctx, univs
+      UVars.UContext.instance uctx, univs
   in
   let cstu = (cst, inst) in
   let inst_type = appvectc (mkConstU cstu) (Context.Rel.instance mkRel 0 params) in
@@ -839,9 +896,7 @@ let declare_class_constant ~univs paramimpls params data =
   Classes.set_typeclass_transparency ~locality:Hints.SuperGlobal
     [Tacred.EvalConstRef cst] false;
   let () =
-    let csb = Global.lookup_constant cst in
-    let poly = Declareops.constant_is_polymorphic csb in
-    declare_proj_coercion_instance ~flags:proj_flags (GlobRef.ConstRef proj_cst) cref ~poly ~with_coercion:false in
+    declare_proj_coercion_instance ~flags:proj_flags (GlobRef.ConstRef proj_cst) cref ~with_coercion:false in
   let m = {
     meth_name = Name proj_name;
     meth_info = None;
@@ -892,14 +947,14 @@ let declare_class ~univs params inds def data =
   let univs, params, fields =
     match fst univs with
     | UState.Polymorphic_entry uctx ->
-      let usubst, auctx = Univ.abstract_universes uctx in
-      let usubst = Univ.make_instance_subst usubst in
+      let usubst, auctx = UVars.abstract_universes uctx in
+      let usubst = UVars.make_instance_subst usubst in
       let map c = Vars.subst_univs_level_constr usubst c in
       let fields = Context.Rel.map map fields in
       let params = Context.Rel.map map params in
       auctx, params, fields
     | UState.Monomorphic_entry _ ->
-      Univ.AbstractContext.empty, params, fields
+      UVars.AbstractContext.empty, params, fields
   in
   let map (impl, projs) =
     let k =
@@ -919,9 +974,9 @@ let add_constant_class cst =
   let env = Global.env () in
   let ty, univs = Typeops.type_of_global_in_context env (GlobRef.ConstRef cst) in
   let r = (Environ.lookup_constant cst env).const_relevance in
-  let ctx, _ = decompose_prod_assum ty in
+  let ctx, _ = decompose_prod_decls ty in
   let args = Context.Rel.instance Constr.mkRel 0 ctx in
-  let t = mkApp (mkConstU (cst, Univ.make_abstract_instance univs), args) in
+  let t = mkApp (mkConstU (cst, UVars.make_abstract_instance univs), args) in
   let tc =
     { cl_univs = univs;
       cl_impl = GlobRef.ConstRef cst;
@@ -942,7 +997,7 @@ let add_inductive_class ind =
   let k =
     let ctx = oneind.mind_arity_ctxt in
     let univs = Declareops.inductive_polymorphic_context mind in
-    let inst = Univ.make_abstract_instance univs in
+    let inst = UVars.make_abstract_instance univs in
     let ty = Inductive.type_of_inductive ((mind, oneind), inst) in
     let r = oneind.mind_relevance in
       { cl_univs = univs;
@@ -956,7 +1011,7 @@ let add_inductive_class ind =
   Classes.add_class k
 
 let warn_already_existing_class =
-  CWarnings.create ~name:"already-existing-class" ~category:"automation" Pp.(fun g ->
+  CWarnings.create ~name:"already-existing-class" ~category:CWarnings.CoreCategories.automation Pp.(fun g ->
       Printer.pr_global g ++ str " is already declared as a typeclass.")
 
 let declare_existing_class g =
@@ -974,7 +1029,7 @@ let declare_existing_class g =
 
 let definition_structure udecl kind ~template ~cumulative ~poly ~primitive_proj
     finite (records : Ast.t list) : GlobRef.t list =
-  let impargs, params, univs, variances, projections_kind, data =
+  let impargs, params, univs, variances, projections_kind, data, indlocs =
     let definitional = kind_class kind = DefClass in
     pre_process_structure ~definitional udecl kind ~poly records
   in
@@ -985,7 +1040,10 @@ let definition_structure udecl kind ~template ~cumulative ~poly ~primitive_proj
          (started in 8.16, c.f., https://github.com/coq/coq/pull/15802 ) *)
       let data = if kind_class kind = NotClass then data else
           List.map (fun d -> { d with Data.is_coercion = false }) data in
-      let structure = interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind data in
+      let structure =
+        interp_structure_core
+          ~cumulative finite ~univs ~variances ~primitive_proj
+          impargs params template ~projections_kind ~indlocs data in
       declare_structure structure
   in
   if kind_class kind <> NotClass then declare_class ~univs params inds def data;
