@@ -69,6 +69,12 @@ module DefAttributes = struct
 
   let clearbody = bool_attribute ~name:"clearbody"
 
+  let make_information deprecated =
+    let open Library_info in
+    match deprecated with
+    | Some depr -> [Deprecation depr]
+    | None -> []
+
   let parse ?(coercion=false) ?(discharge=NoDischarge) f =
     let clearbody = match discharge with DoDischarge -> clearbody | NoDischarge -> return None in
     let (((((((locality, deprecated), polymorphic), program),
@@ -443,7 +449,7 @@ let print_universes ?loc ~sort ~subgraph dst =
     if Global.is_joined_environment () then mt ()
     else str"There may remain asynchronous universe constraints"
   in
-  let prl = UnivNames.pr_with_global_universes in
+  let prl = UnivNames.pr_level_with_global_universes in
   begin match dst with
     | None -> UGraph.pr_universes prl univ ++ pr_remaining
     | Some s -> dump_universes_gen (fun u -> Pp.string_of_ppcmds (prl u)) univ s
@@ -570,25 +576,20 @@ let program_inference_hook env sigma ev =
     user_err Pp.(str "The statement obligations could not be resolved \
                       automatically, write a statement definition first.")
 
-let vernac_set_used_variables ~pstate using : Declare.Proof.t =
+let vernac_set_used_variables pstate using : Declare.Proof.t =
   let env = Global.env () in
   let sigma, _ = Declare.Proof.get_current_context pstate in
+  let fixnames = Declare.Proof.get_recnames pstate in
   let initial_goals pf = Proofview.initial_goals Proof.((data pf).entry) in
   let terms = List.map pi3 (initial_goals (Declare.Proof.get pstate)) in
-  let using = Proof_using.definition_using env sigma ~using ~terms in
-  let vars = Environ.named_context env in
-  Names.Id.Set.iter (fun id ->
-      if not (List.exists (NamedDecl.get_id %> Id.equal id) vars) then
-        user_err
-          (str "Unknown variable: " ++ Id.print id ++ str "."))
-    using;
+  let using = Proof_using.definition_using env sigma ~fixnames ~using ~terms in
   let _, pstate = Declare.Proof.set_used_variables pstate ~using in
   pstate
 
 let vernac_set_used_variables_opt ?using pstate =
   match using with
   | None -> pstate
-  | Some expr -> vernac_set_used_variables ~pstate expr
+  | Some expr -> vernac_set_used_variables pstate expr
 
 (* XXX: Interpretation of lemma command, duplication with ComFixpoint
    / ComDefinition ? *)
@@ -645,7 +646,7 @@ let start_lemma_com ~typing_flags ~program_mode ~poly ~scope ?clearbody ~kind ~d
 
 let vernac_definition_hook ~canonical_instance ~local ~poly ~reversible = let open Decls in function
 | Coercion ->
-  Some (ComCoercion.add_coercion_hook ~poly ~reversible)
+  Some (ComCoercion.add_coercion_hook ~reversible)
 | CanonicalStructure ->
   Some (Declare.Hook.(make (fun { S.dref } -> Canonical.declare_canonical_structure ?local dref)))
 | SubClass ->
@@ -1368,7 +1369,7 @@ let vernac_end_segment ~pm ~proof ({v=id} as lid) =
   | _ -> assert false
 
 let vernac_end_segment lid =
-  Vernacextend.TypedVernac {
+  Vernactypes.TypedVernac {
     inprog = Use; outprog = Pop; inproof = UseOpt; outproof = No;
     run = (fun ~pm ~proof ->
         let () = vernac_end_segment ~pm ~proof lid in
@@ -1376,9 +1377,9 @@ let vernac_end_segment lid =
   }
 
 let vernac_begin_segment ~interactive f =
-  let inproof = Vernacextend.InProof.(if interactive then Reject else Ignore) in
-  let outprog = Vernacextend.OutProg.(if interactive then Push else No) in
-  Vernacextend.TypedVernac {
+  let inproof = Vernactypes.InProof.(if interactive then Reject else Ignore) in
+  let outprog = Vernactypes.OutProg.(if interactive then Push else No) in
+  Vernactypes.TypedVernac {
     inprog = Ignore; outprog; inproof; outproof = No;
     run = (fun ~pm ~proof ->
         let () = f () in
@@ -1404,7 +1405,9 @@ let vernac_require_interp needed modrefl export qidl =
   in
   if Dumpglob.dump () then
     List.iter2 (fun ({CAst.loc},_) dp -> Dumpglob.dump_libref ?loc dp "lib") qidl modrefl;
+  (* Load *)
   Library.require_library_from_dirpath needed;
+  (* Import*)
   Option.iter (fun (export,cats) ->
       let cats = interp_import_cats cats in
       List.iter2 (fun m (_,f) ->
@@ -1425,13 +1428,13 @@ let vernac_coercion ~atts ref qidst =
   let ref' = smart_global ref in
   match qidst with
   | Some (qids, qidt) ->
-     let (local, poly), reversible =
-       Attributes.parse Notations.(locality ++ polymorphic ++ reversible) atts in
+     let local, reversible =
+       Attributes.parse Notations.(locality ++ reversible) atts in
      let local = enforce_locality local in
      let reversible = Option.default false reversible in
      let target = cl_of_qualid qidt in
      let source = cl_of_qualid qids in
-     ComCoercion.try_add_new_coercion_with_target ref' ~local ~poly ~reversible
+     ComCoercion.try_add_new_coercion_with_target ref' ~local ~reversible
        ~source ~target;
      Flags.if_verbose Feedback.msg_info (pr_global ref' ++ str " is now a coercion")
   | None ->
@@ -1496,7 +1499,7 @@ let vernac_existing_class id =
 let command_focus = Proof.new_focus_kind ()
 let focus_command_cond = Proof.no_cond command_focus
 
-let vernac_set_end_tac ~pstate tac =
+let vernac_set_end_tac pstate tac =
   let env = Genintern.empty_glob_sign ~strict:true (Global.env ()) in
   let _, tac = Genintern.generic_intern env tac in
   (* TO DO verifier s'il faut pas mettre exist s | TacId s ici*)
@@ -1919,11 +1922,6 @@ let vernac_global_check c =
   pr_universe_ctx_set sigma uctx
 
 
-let get_nth_goal ~pstate n =
-  let pf = Declare.Proof.get pstate in
-  let Proof.{goals;sigma} = Proof.data pf in
-  (sigma, List.nth goals (n - 1))
-
 (* Printing "About" information of a hypothesis of the current goal.
    We only print the type and a small statement to this comes from the
    goal. Precondition: there must be at least one current goal. *)
@@ -1938,16 +1936,20 @@ let print_about_hyp_globs ~pstate ?loc ref_or_by_not udecl glopt =
     in
     (* FIXME error on non None udecl if we find the hyp. *)
     let glnumopt = query_command_selector ?loc glopt in
-    let (sigma, ev), id =
+    let pf = Declare.Proof.get pstate in
+    let Proof.{goals; sigma} = Proof.data pf in
+    let ev, id =
       let open Constrexpr in
       match glnumopt, ref_or_by_not.v with
       | None,AN qid when qualid_is_ident qid -> (* goal number not given, catch any failure *)
-         (try get_nth_goal ~pstate 1, qualid_basename qid with _ -> raise NoHyp)
+        (match List.nth_opt goals 0 with
+         | None -> raise NoHyp
+         | Some goal -> goal), qualid_basename qid
       | Some n,AN qid when qualid_is_ident qid ->  (* goal number given, catch if wong *)
-         (try get_nth_goal ~pstate n, qualid_basename qid
-          with
-            Failure _ -> user_err ?loc
-                          (str "No such goal: " ++ int n ++ str "."))
+        (match List.nth_opt goals (n - 1) with
+         | None  -> user_err ?loc
+                      (str "No such goal: " ++ int n ++ str ".")
+         | Some goal -> goal), qualid_basename qid
       | _ , _ -> raise NoHyp in
     let hyps = Evd.evar_filtered_context (Evd.find_undefined sigma ev) in
     let decl = Context.Named.lookup id hyps in
@@ -2082,6 +2084,13 @@ let vernac_register qid r =
     end
     else Coqlib.register_ref (Libnames.string_of_qualid n) gr
 
+let vernac_library_attributes atts =
+  if Global.is_curmod_library () && not (Lib.sections_are_opened ()) then
+    let deprecated = Attributes.parse deprecation atts in
+    Lib.Synterp.declare_info (DefAttributes.make_information deprecated)
+  else
+    user_err (Pp.str "A library attribute should be at toplevel of the library.")
+
 (********************)
 (* Proof management *)
 
@@ -2214,7 +2223,26 @@ let vernac_validate_proof ~pstate =
     str "No issues found."
   else prlist_with_sep fnl snd (Evar.Map.bindings evar_issues)
 
-let translate_vernac_synterp ?loc ~atts v = let open Vernacextend in match v with
+let vernac_proof pstate tac using =
+  let is_let = match Declare.Proof.definition_scope pstate with
+    | Discharge -> true
+    | Global _ -> false
+  in
+  let using = if not is_let then Option.append using (Proof_using.get_default_proof_using ())
+    else
+      let () = if Option.has_some using
+        then CErrors.user_err Pp.(str "Let does not support Proof using.")
+      in
+      None
+  in
+  let tacs = if Option.is_empty tac then "tac:no" else "tac:yes" in
+  let usings = if Option.is_empty using then "using:no" else "using:yes" in
+  Aux_file.record_in_aux_at "VernacProof" (tacs^" "^usings);
+  let pstate = Option.fold_left vernac_set_end_tac pstate tac in
+  let pstate = Option.fold_left vernac_set_used_variables pstate using in
+  pstate
+
+let translate_vernac_synterp ?loc ~atts v = let open Vernactypes in match v with
   | EVernacNotation { local; decl } ->
     vtdefault(fun () -> Metasyntax.add_notation_interpretation ~local (Global.env()) decl)
 
@@ -2274,7 +2302,7 @@ let translate_vernac_synterp ?loc ~atts v = let open Vernacextend in match v wit
   (* Extensions *)
   | EVernacExtend f -> f
 
-let translate_pure_vernac ?loc ~atts v = let open Vernacextend in match v with
+let translate_pure_vernac ?loc ~atts v = let open Vernactypes in match v with
   | VernacAbortAll
   | VernacRestart
   | VernacUndo _
@@ -2505,6 +2533,10 @@ let translate_pure_vernac ?loc ~atts v = let open Vernacextend in match v with
         unsupported_attributes atts;
         Flags.if_verbose Feedback.msg_info (str "Comments ok\n"))
 
+  | VernacAttributes atts ->
+    vtdefault(fun () ->
+        vernac_library_attributes atts)
+
   (* Proof management *)
   | VernacFocus n ->
     vtmodifyproof(unsupported_attributes atts;vernac_focus n)
@@ -2541,12 +2573,7 @@ let translate_pure_vernac ?loc ~atts v = let open Vernacextend in match v with
   | VernacProof (tac, using) ->
     vtmodifyproof(fun ~pstate ->
     unsupported_attributes atts;
-    let using = Option.append using (Proof_using.get_default_proof_using ()) in
-    let tacs = if Option.is_empty tac then "tac:no" else "tac:yes" in
-    let usings = if Option.is_empty using then "using:no" else "using:yes" in
-    Aux_file.record_in_aux_at "VernacProof" (tacs^" "^usings);
-    let pstate = Option.cata (vernac_set_end_tac ~pstate) pstate tac in
-    Option.cata (vernac_set_used_variables ~pstate) pstate using)
+    vernac_proof pstate tac using)
 
   | VernacEndProof pe ->
     unsupported_attributes atts;
